setfenv(1, DockingStation)

local gsub, ipairs, strtrim, tonumber, tostring, type = gsub, ipairs, strtrim, tonumber, tostring, type

local _field, _type

--[[-----------------------------------------------------------------------------
Methods to verify a value is of a designated type. Verification is done per field
unless the field being verified is 'type' then the entire dataObject is passed
for verification.
-------------------------------------------------------------------------------]]
local function verify_type(value)
	if type(value) == _type then
		return value
	end
end

local verify
verify = setmetatable({
	["color"] = function(value)
		if type(value) ~= 'number' or value < 0 or value > 1 then return end
		return value
	end,

	["coords"] = function(value)
		if type(value) ~= 'table' or not (#value == 4 or #value == 8) then return end
		for _, coord in ipairs(value) do
			if type(tonumber(coord)) ~= 'number' then return end
		end
		return value
	end,

	["frame"] = function(value)
		if not IsFrame(value) then return end
		return value
	end,

	['function'] = function(value)
		if type(value) == 'function' then
			return value
		elseif type(value) == 'table' then
			local mt = getmetatable(value)
			if mt and type(mt) == 'table' and type(mt.__call) == 'function' then
				return value
			end
		end
	end,

	["icon"] = function(value)
		if type(value) == 'string' then
			return value
		elseif type(value) == 'table' then
			if #value < 3 or #value > 4 then return end										-- Allow alpha but it will not be used
			for index = 1, 3 do
				if not verify["color"](value[index]) then return end
			end
			return value
		end
	end,

	['string'] = function(value)
		if value == nil then
			return ""
		else
			value = strtrim(gsub(tostring(value), "[%c]+", " "), nil)					-- strtrim() can take more than one arg so the nil overrides it
			return value
		end
	end
}, { __index = function(self, index) _type = index return verify_type end })

--[[-----------------------------------------------------------------------------
Supported fields for each dataObject.type and their designated types. Any field
not listed for a dataObject will not be updated for its plugin. Also, if a value
doesn't pass verification then it won't be updated in the plugin.  Setting a
field to false will generate a warning if it is used with that plugin type.
-------------------------------------------------------------------------------]]
local supported = {
	['data source'] = {
		icon = "icon",
		iconB = "color",
		iconCoords = "coords",
		iconG = "color",
		iconR = "color",
		keyBind = 'string',																			-- Not part of standard
		label = 'string',
		OnClick = 'function',
		OnEnter = 'function',
		OnLeave = 'function',
		OnTooltipShow = 'function',
		suffix = 'string',
		text = 'string',
		texcoord = "coords",																			-- Not part of standard (alias for 'iconCoords')
		tooltip = "frame",
		tooltiptext = 'string',																		-- No longer part of standard
		value = 'string'
	},
--[[
	['group'] = {
	},
]]
	['launcher'] = {
		icon = "icon",
		iconB = "color",
		iconCoords = "coords",
		iconG = "color",
		iconR = "color",
		keyBind = 'string',																			-- Not part of standard
		label = 'string',
		OnClick = 'function',
		OnEnter = 'function',
		OnLeave = 'function',
		OnTooltipShow = 'function',
		suffix = false,
		text = false,
		texcoord = "coords",																			-- Not part of standard (alias for 'iconCoords')
		tooltip = "frame",
		tooltiptext = 'string',																		-- No longer part of standard
		value = false
	},
--[[
	['plugin'] = {
		group = 'string',
		label = 'string'
	}
]]
}

--[[-----------------------------------------------------------------------------
Methods to update plugin data with minimal calls to plugin:Refresh(). Only
called after UpdateData has verified that the field and value are ok to be
updated.
-------------------------------------------------------------------------------]]
local function update_field(plugin, value)
	plugin.data[_field] = value
end

local update = setmetatable({
	['group'] = function(plugin, value)
		plugin.settings.group = value
		PluginList:Add(plugin.name)
	end,

	['icon'] = function(plugin, value)
		local data, icon = plugin.data, plugin.icon
		data.icon = value
		SetTexture(icon, value)
		if data.iconCoords or plugin.settings.iconZoom then
			SetTexCoord(icon, data.iconCoords, plugin.settings.iconZoom)
		end
		if not plugin.settings.iconHide then
			if icon:IsShown() then
				if value == nil then
					icon:Hide()
					return true
				end
			elseif value then
				icon:Show()
				return true
			end
		end
	end,

	['iconB'] = function(plugin, value, field)
		local data = plugin.data
		data[field] = value
		if plugin.icon then
			plugin.icon:SetVertexColor(data.iconR or 0, data.iconG or 0, data.iconB or 0, 1)
		end
	end,

	['iconCoords'] = function(plugin, value)
		plugin.data.iconCoords = value
		SetTexCoord(plugin.icon, value, plugin.settings.iconZoom)
		plugin:Refresh()																				-- Temp fix until size is updated directly
	end,

	['label'] = function(plugin, value)
		if value == "" then
			value = strtrim(plugin.name)
		end
		plugin.data.label, plugin.settings.label = value, value
		PluginList:Add(plugin.name)
		UpdatePluginConfig()
		if not plugin.settings.labelHide or not (plugin.icon:IsShown() or plugin.text:IsShown()) then
			return true
		end
	end,

	['suffix'] = function(plugin, value)
		plugin.data.suffix = value ~= "" and value or nil
		if plugin.data.value and not plugin.settings.suffixHide then
			return true
		end
	end,

	['text'] = function(plugin, value)
		plugin.data.text = value ~= "" and value or nil
		if not plugin.data.value and not plugin.settings.textHide then
			return true
		end
	end,

	['value'] = function(plugin, value)
		plugin.data.value = value ~= "" and value or nil
		if not plugin.settings.textHide then
			return true
		end
	end
}, { __index = function(self, index) _field = index return update_field end })
update['iconG'], update['iconR'], update['texcoord'] = update['iconB'], update['iconB'], update['iconCoords']

--[[-----------------------------------------------------------------------------
Helpers
-------------------------------------------------------------------------------]]
local reportIssued = { }
local function report(message, isError)
	if not settings.report or reportIssued[message] then return end
	reportIssued[message] = true
	local status
	if isError then
		status = "|cffff1e1e[ERROR]"
	else
		status = "|cffffff1e[WARNING]"
	end
	DEFAULT_CHAT_FRAME:AddMessage("|cff1eff1e[" .. title .. "] " .. status .. "|r " .. message)
end

local function wait_for_type(_, name, _, _, dataObject)
	ValidateDataObject(_, name, dataObject)
	LDB.UnregisterCallback(addon, 'LibDataBroker_AttributeChanged_' .. name .. '_launcher')
	LDB.UnregisterCallback(addon, 'LibDataBroker_AttributeChanged_' .. name .. '_type')
end

--[[-----------------------------------------------------------------------------
Global to addon
-------------------------------------------------------------------------------]]
function UpdateDataObject(_, name, field, value)
	local plugin = plugins[name]
	local type = supported[plugin.data.type][field]
	if type then
		local newValue = verify[type](value)
		if newValue then
			if update[field](plugin, newValue, field) then
				plugin:Refresh(true)
			end
		elseif value ~= nil then
			report("Plugin '" .. tostring(name) .. "' attempted to change field '" .. tostring(field) .. "' to an inappropriate value '" .. tostring(value) .. "'", true)
		end
	elseif type == false then
		report("Plugin '" .. tostring(name) .. "' is using field '" .. tostring(field) .. "' which is not supported for a plugin of type '" .. tostring(plugin.data.type) .. "'")
	end
end

function ValidateDataObject(_, name, dataObject)
	local type = dataObject.type
	if type == nil then																				-- Just in case type == false
		if dataObject.launcher then
			type = dataObject.launcher
		else
			LDB.RegisterCallback(addon, 'LibDataBroker_AttributeChanged_' .. name .. '_launcher', wait_for_type)
			LDB.RegisterCallback(addon, 'LibDataBroker_AttributeChanged_' .. name .. '_type', wait_for_type)
			return
		end
	end
	if plugins[name] then																			-- Is this even possible?
		report("Plugin '" .. tostring(name) .. "' already exists", true) 
	elseif not supported[type] then
		report("Data object of type '" .. tostring(type) .. "' was detected but is not supported")
	else
		CreatePlugin(name, dataObject, type)
	end
end
