setfenv(1, DockingStation)

local format, gsub, ipairs, pairs, select, strlower, unpack, wipe = format, gsub, ipairs, pairs, select, strlower, unpack, wipe
local GetMouseFocus, UIParent = GetMouseFocus, UIParent

panels = { }
local panels = panels

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local anchors, backdropInsets = { Center = 'LEFT', Left = 'LEFT', Right = 'RIGHT' }, { left = 0, right = 0, top = 0, bottom = 0 }
local backdrop = { bgFile = "", tile = false, tileSize = 0, edgeFile = "", edgeSize = 0, insets = backdropInsets }
local connections, directions = { LEFT = 'RIGHT', RIGHT = 'LEFT' }, { Center = 1, Left = 1, Right = -1 }
local jostle_regions = { BOTTOM = 'BOTTOM', BOTTOMLEFT = 'BOTTOM', BOTTOMRIGHT = 'BOTTOM', TOP = 'TOP', TOPLEFT = 'TOP', TOPRIGHT = 'TOP' }
local recycledPanels = { }

local AnchorOffsets = {
	['BOTTOM'] = function(frame)
		return frame:GetCenter() - SCREEN_WIDTH_MIDDLE, frame:GetBottom() - UIParent:GetBottom()
	end,
	['BOTTOMLEFT'] = function(frame)
		return frame:GetLeft() - UIParent:GetLeft(), frame:GetBottom() - UIParent:GetBottom()
	end,
	['BOTTOMRIGHT'] = function(frame)
		return frame:GetRight() - UIParent:GetRight(), frame:GetBottom() - UIParent:GetBottom()
	end,
	['CENTER'] = function(frame)
		local x, y = frame:GetCenter()
		return x - SCREEN_WIDTH_MIDDLE, y - SCREEN_HEIGHT_MIDDLE
	end,
	['LEFT'] = function(frame)
		local _, y = frame:GetCenter()
		return frame:GetLeft() - UIParent:GetLeft(), y - SCREEN_HEIGHT_MIDDLE
	end,
	['RIGHT'] = function(frame)
		local _, y = frame:GetCenter()
		return frame:GetRight() - UIParent:GetRight(), y - SCREEN_HEIGHT_MIDDLE
	end,
	['TOP'] = function(frame)
		return frame:GetCenter() - SCREEN_WIDTH_MIDDLE, frame:GetTop() - UIParent:GetTop()
	end,
	['TOPLEFT'] = function(frame)
		return frame:GetLeft() - UIParent:GetLeft(), frame:GetTop() - UIParent:GetTop()
	end,
	['TOPRIGHT'] = function(frame)
		return frame:GetRight() - UIParent:GetRight(), frame:GetTop() - UIParent:GetTop()
	end
}

local FlipFlop = {
	[false] = {	[false] = { 0, 1, 0, 1 }, [true] = { 0, 1, 1, 0 } },
	[true] = { [false] = { 1, 0, 0, 1 }, [true] = { 1, 0, 1, 0 } }
}

local function UpdateOrder()
	local level, panel
	for index, ID in ipairs(settings.panelOrder) do
		panel = panels[ID]
		if panel then
			level = index * 16
			panel:SetFrameLevel(level)
			panel.auxFrame:SetFrameLevel(level)
			level = level + 1
			for sectionType in pairs(section_types) do
				for _, name in ipairs(panel.settings[sectionType]) do
					if plugins[name] then
						plugins[name]:SetFrameLevel(level)
					end
				end
			end
		end
	end
end

local methods = {
	["AddPlugin"] = function(self, name, section, index)
		local removePanel, removeSection, removeIndex = GetPluginLocation(name)
		if removePanel then
			tremove(removePanel.settings[removeSection], removeIndex)
			removePanel:Refresh()
		end
		section = self.settings[section or "Left"]
		tinsert(section, index or #section + 1, name)
		if plugins[name] then
			plugins[name]:SetState()
		else
			SetInheritance(name, self)
		end
		UpdatePluginConfig()
	end,

	["Delete"] = function(self)
		local id, panelOrder = self.id, settings.panelOrder
		for sectionType in pairs(section_types) do
			for _, name in ipairs(self.settings[sectionType]) do
				if settings.plugins[name] then
					settings.plugins[name].enable = false
				end
				if plugins[name] then
					plugins[name]:SetState(true)
				else
					SetInheritance(name, nil)
				end
		    end
		end
		tremove(settings.panels, id)
		self:Recycle()
		for ID = id, #panels do
			panels[ID].id = ID
		end
		tremove_byVal(panelOrder, id)
		for index, ID in ipairs(panelOrder) do
			if ID > id then
				panelOrder[index] = panelOrder[index] - 1
			end
		end
		UpdateOrder()
		PanelList:Delete(id)
	end,

	["Disable"] = function(self)
		self.settings.enable = false
		if self.jostle then
			LJ:Unregister(self)
			self.jostle = nil
		end
		self:Hide()
	end,

	["Enable"] = function(self)
		self.settings.enable = true
		self:Refresh()
		self:Show()
	end,

	["GetLabel"] = function(self, forSorting)
		local label = self.settings.template
		if not forSorting then
			label = gsub(label, "$id", self.id)
			label = gsub(label, "$#", #panels)
		else
			label = strlower(RemoveColorCodes(label))
			label = gsub(label, "$id", format("%03d", self.id))
			label = gsub(label, "$#", format("%03d", #panels))
		end
		return label
	end,

	["GetOffset"] = function(self)
		local x, y = AnchorOffsets[self.anchor](self)
		return x / SCREEN_WIDTH, y / SCREEN_HEIGHT
	end,

	["GetOrder"] = function(self)
		local id, panelOrder = self.id, settings.panelOrder
		for index = 1, #panelOrder do
			if panelOrder[index] == id then
				return index
			end
		end
	end,

	["HasPlugin"] = function(self, name)
		for section in pairs(section_types) do
			for index, pluginName in ipairs(self.settings[section]) do
				if pluginName == name then
					return section, index
				end
			end
		end
	end,

	["Recycle"] = function(self)
		self:Hide()
		if self.jostle then
			LJ:Unregister(self)
		end
		tremove(panels, self.id)																	-- Repercussions of this are handled by the caller (if needed)
		self.auxFrame, self.overlay, self.Center, self.Left, self.Right, self._SetWidth, self[0], self = self.auxFrame, self.overlay, wipe(self.Center), wipe(self.Left), wipe(self.Right), self._SetWidth, self[0], wipe(self)
		recycledPanels[#recycledPanels + 1] = self
	end,

	["Refresh"] = function(self)
		local settings = self.settings
		if not settings.enable then return end
		local strata = strata_layers[settings.strata]
		self.anchor, self.height = settings.anchor, settings.height * SCREEN_HEIGHT
		self:SetFrameStrata(strata)
		self:UpdateVisibility()

		local clamp, x, y = settings.screenClamp, self:GetOffset()
		self:SetClampedToScreen(false)
		self:SetClampRectInsets(clamp, -clamp, -clamp, clamp)
		self:SetHeight(self.height)
		self:SetWidth(settings.width * SCREEN_WIDTH)
		self:SetClampedToScreen(true)
		self:SetOffset(x, y)

		local bgInset = settings.bgInset
		backdrop.bgFile = LSM:Fetch('statusbar', settings.bgTexture) or LSM:Fetch('statusbar', "Solid")
		backdrop.edgeFile = LSM:Fetch('border', settings.borderTexture) or LSM:Fetch('border', 'None')
		backdrop.edgeSize = settings.borderSize
		for inset in pairs(backdropInsets) do
			backdropInsets[inset] = bgInset
		end
		self:SetBackdrop(backdrop)
		self:SetBackdropColor(settings.bgColorR, settings.bgColorG, settings.bgColorB, settings.bgColorA)
		self:SetBackdropBorderColor(settings.borderColorR, settings.borderColorG, settings.borderColorB, settings.borderColorA)

		local overlay = self.overlay
		overlay:SetTexture(LSM:Fetch('overlay', settings.bgOverlay) or LSM:Fetch('overlay', "None"))
		overlay:SetTexCoord(unpack(FlipFlop[settings.bgOverlayFlip][settings.bgOverlayFlop]))
		overlay:SetVertexColor(settings.bgOverlayColorR, settings.bgOverlayColorG, settings.bgOverlayColorB, settings.bgOverlayColorA)

		local anchor, anchorTo, connection, frame, offset, plugin, section, spacing
		for sectionType in pairs(section_types) do
			anchor, section = anchors[sectionType], wipe(self[sectionType])
			anchorTo, connection = anchor, connections[anchor]
			spacing = settings["spacing" .. sectionType] * directions[sectionType]
			if sectionType ~= "Center" then
				frame, offset = self, settings["spacing" .. sectionType .. "Edge"] * directions[sectionType]
			else
				frame, offset = self.auxFrame, 0
			end
			for _, name in ipairs(settings[sectionType]) do
				plugin = plugins[name]
				if plugin then
					if plugin.settings.enable then
						section[#section + 1] = plugin
						plugin:SetFrameStrata(strata)
						plugin:ClearAllPoints()
						plugin:SetPoint(anchor, frame, anchorTo, offset, 0)
						plugin:Refresh()
						anchorTo, frame, offset = connection, plugin, spacing
					end
				end
			end
		end
	end,

	["SaveOffset"] = function(self)
		local settings = self.settings
		settings.offsetX, settings.offsetY = self:GetOffset()
		if settings.moveBlizzard then
			local region = jostle_regions[self.anchor] or (select(2, self:GetCenter()) < SCREEN_HEIGHT_MIDDLE and 'BOTTOM') or 'TOP'
			if region ~= self.jostle then
				if region == 'TOP' then
					LJ:RegisterTop(self)
				else
					LJ:RegisterBottom(self)
				end
				self.jostle = region
			else
				LJ.frame:Schedule()
			end
		elseif self.jostle then
			LJ:Unregister(self)
			self.jostle = nil
		end
	end,

	["SetOffset"] = function(self, x, y)
		self:ClearAllPoints()
		self:SetPoint(self.anchor, UIParent, self.anchor, x * SCREEN_WIDTH, y * SCREEN_HEIGHT)
		self:SaveOffset()
	end,

	["UpdateOrder"] = function(self)
		tremove_byVal(settings.panelOrder, self.id)
		settings.panelOrder[#settings.panelOrder + 1] = self.id
		UpdateOrder()
	end,

	["UpdateVisibility"] = function(self)
		local settings = self.settings
		if (CONFIG_IS_OPEN and CONFIG_IS_OPEN ~= self.id) or DRAG_IN_PROGRESS then
			self:SetAlpha(1)
		elseif settings.alphaCombat and PLAYER_IN_COMBAT then
			self:SetAlpha(settings.alphaFaded)
		elseif settings.alphaMouse then
			local mouseFocus = TOOLTIP_ON_PANEL or GetMouseFocus() or UIParent
			if mouseFocus == self or mouseFocus.panel == self then
				self:SetAlpha(settings.alphaNormal > settings.alphaFaded and settings.alphaNormal or settings.alphaFaded)
			else
				self:SetAlpha(settings.alphaFaded < settings.alphaNormal and settings.alphaFaded or settings.alphaNormal)
			end
		else
			self:SetAlpha(settings.alphaNormal)
		end
	end
}

local mt_panel = { __index = setmetatable(methods, MT_FRAME) }

--[[-----------------------------------------------------------------------------
Scripts
-------------------------------------------------------------------------------]]
local OnUpdate
do
	local x, y
	function OnUpdate(self)
		x, y = self:GetCenter()
		if self.x ~= x or self.y ~= y then
			self:SetScript('OnUpdate', nil)
			self:UpdateOrder()
		end
	end
end

local function OnEnter(self)
	self:UpdateVisibility()
	if not CONFIG_IS_OPEN then return end
	StartTooltipDetection()
	GameTooltip:SetOwner(self, 'ANCHOR_NONE')
	GameTooltip:SetText(self:GetLabel())
	StopTooltipDetection()
end

local function OnLeave(self)
	self:UpdateVisibility()
	if CONFIG_IS_OPEN then
		GameTooltip:Hide()
	end
	ReleaseTooltip()
end

local function OnMouseDown(self, button)
	if self.settings.lockPanel or button ~= 'LeftButton' then return end
	OnLeave(self)
	self.x, self.y = self:GetCenter()
	self:SetScript('OnUpdate', OnUpdate)
	self:SetFrameStrata('TOOLTIP')
	self:StartMoving()
	self.moving = true
end

local function OnMouseUp(self, button)
	if button == 'LeftButton' and self.moving then
		self.moving = nil
		self:StopMovingOrSizing()
		self:SetFrameStrata(strata_layers[self.settings.strata])
		self:SetScript('OnUpdate', nil)
		self:SaveOffset()
		OnEnter(self)
	elseif button == 'RightButton' and GetMouseFocus() == self and self.settings.rightClickConfig and CONFIG_IS_OPEN ~= self.id then
		ConfigFrames[1]()																				-- Do it this way to force loading of config module
		PanelList:Select(self.id)
		CONFIG_IS_OPEN = self.id
		AllPanels("UpdateVisibility")
	end
end

--[[-----------------------------------------------------------------------------
Support for detecting and processing combat states
-------------------------------------------------------------------------------]]
local processCombatState = CreateFrame('Frame', nil, UIParent)
processCombatState:SetScript('OnEvent', function(self, event)
	PLAYER_IN_COMBAT = event == 'PLAYER_REGEN_DISABLED'
	AllPanels("UpdateVisibility")
end)
processCombatState:RegisterEvent('PLAYER_REGEN_DISABLED')
processCombatState:RegisterEvent('PLAYER_REGEN_ENABLED')

--[[-----------------------------------------------------------------------------
Support for centering plugins in a panel's center section
-------------------------------------------------------------------------------]]
local delayCentering, needsCentered, _SetWidth = CreateFrame('Frame', nil, UIParent), { }, utilityFrame.SetWidth
delayCentering:SetScript('OnUpdate', function(self)
	for frame, plugins in pairs(needsCentered) do
		frame:SetWidth(plugins[#plugins]:GetRight() - plugins[1]:GetLeft())
	end
	wipe(needsCentered)
	self:Hide()
end)
delayCentering:Hide()

--[[-----------------------------------------------------------------------------
Global to addon
-------------------------------------------------------------------------------]]
PanelList, PLAYER_IN_COMBAT = { Add = DoNothing, Delete = DoNothing, Refresh = DoNothing, Select = DoNothing }, false

function CreatePanel(id)
	id = id or #panels + 1
	local panel = panels[id]
	if not panel then
		panel = recycledPanels[#recycledPanels]
		if panel then
			recycledPanels[#recycledPanels] = nil
			panel:ClearAllPoints()
			panel:Show()
		else
			panel = CreateFrame('Frame', nil, UIParent)
			panel:EnableMouse(true)
			panel:SetMovable(true)
			panel:SetScript('OnEnter', OnEnter)
			panel:SetScript('OnLeave', OnLeave)
			panel:SetScript('OnMouseDown', OnMouseDown)
			panel:SetScript('OnMouseUp', OnMouseUp)
			local overlay = panel:CreateTexture(nil, 'OVERLAY')
			overlay:SetAllPoints(panel)
			panel.overlay = overlay
			local auxFrame, center = CreateFrame('Frame', nil, panel), { }
			auxFrame:SetPoint('CENTER', panel, 'CENTER')
			auxFrame:SetHeight(1)
			auxFrame:SetWidth(1)
			panel._SetWidth = function(plugin, value)											-- New SetWidth for plugins in this panel's center section
				_SetWidth(plugin, value)
				needsCentered[auxFrame] = center
				delayCentering:Show()
			end
			panel.auxFrame, panel.Center, panel.Left, panel.Right = auxFrame, center, { }, { }
		end
		panels[id] = setmetatable(panel, mt_panel)
		panel.id = id
	end
	panel:SetClampedToScreen(false)
	panel:SetHeight(1)
	panel:SetWidth(1)
	panel.settings = GetPanelSettings(id)
	panel.anchor = panel.settings.anchor
	panel:SetOffset(panel.settings.offsetX, panel.settings.offsetY)
	if panel:GetOrder() then
		UpdateOrder()
	else
		panel:UpdateOrder()
	end
	if panel.settings.enable then
		panel:Show()
	else
		panel:Hide()
	end
	PanelList:Add(id)
	return panel
end
