if not DockingStation then return end
setfenv(1, DockingStation)

local ipairs, max, min, pairs, rawget, select, strlower, strtrim, tinsert, tremove, tostring, wipe = ipairs, max, min, pairs, rawget, select, strlower, strtrim, tinsert, tremove, tostring, wipe

PanelDropDownList = { }

--[[-----------------------------------------------------------------------------
Helpers
-------------------------------------------------------------------------------]]
local MAX_HEIGHT, MAX_WIDTH, MIN_HEIGHT, MIN_WIDTH = RESOLUTION_HEIGHT / 16, RESOLUTION_WIDTH, RESOLUTION_HEIGHT / 64, RESOLUTION_WIDTH / 64
local anchor_points = { BOTTOM = 'BOTTOM', BOTTOMLEFT = 'BOTTOMLEFT', BOTTOMRIGHT = 'BOTTOMRIGHT', CENTER = 'CENTER', LEFT = 'LEFT', RIGHT = 'RIGHT', TOP = 'TOP', TOPLEFT = 'TOPLEFT', TOPRIGHT = 'TOPRIGHT' }
local panels, plugins, PanelList, PluginList = panels, plugins, PanelList, PluginList
local panelIDs, panelList = { }, { }
local unlock_types = { [0] = "Nothing", [1] = "Panels", [2] = "Plugins", [3] = "Panels and Plugins" }

local function update_panel_list()
	wipe(panelIDs)
	wipe(panelList)
	for index = 1, #PanelList do
		panelIDs[index], panelList[index] = PanelList[index].key, PanelList[index].label
	end
end

--[[-----------------------------------------------------------------------------
Common for all options
-------------------------------------------------------------------------------]]
local ACR_CBH, _category, _field, _root, _settings, _temp = LibStub('AceConfigRegistry-3.0').callbacks

local function AnyLocked(field)
	for id = 1, #panels do
		if panels[id].settings[field] then
			return true
		end
	end
end

local function AnyUnlocked(field)
	for id = 1, #panels do
		if not panels[id].settings[field] then
			return true
		end
	end
end

local function GetData(info)
	_category, _field, _root = info[#info - 1], info[#info], info[1]
	if _root == "Plugins" then
		_settings = settings.plugins[PluginList.selection]
	elseif _root == "Panels" then
		_settings = settings.panels[PanelList.selection]
	else
		_settings = settings
	end
end

local function DisableIfDefault(info)
	GetData(info)
	return _settings[_category .. "Defaults"]
end

local function Get(info)
	GetData(info)
	return _settings[_field]
end

local function GetColor(info)
	GetData(info)
	return _settings[_field .. "R"], _settings[_field .. "G"], _settings[_field .. "B"], _settings[_field .. "A"]
end

local function GetString(info)
	GetData(info)
	return _settings[_category .. _field]
end

local function MassClear(category, ...)
	category = category or ""
	for index = 1, select('#', ...) do
		_settings[category .. select(index, ...)] = nil
	end
end

local function MassLoad(category, ...)
	category = category or ""
	for index = 1, select('#', ...) do
		_temp = category .. select(index, ...)
		_settings[_temp] = _settings[_temp]
	end
end

local function RefreshSettings()
	if _root == "Plugins" then
		if plugins[PluginList.selection] then
			plugins[PluginList.selection]:Refresh()
		end
	elseif _root == "Panels" then
		panels[PanelList.selection]:Refresh()
	else
		AllPanels("Refresh")
	end
end

local function Set(info, value)
	GetData(info)
	_settings[_field] = value or false
end

local function SetColor(info, red, green, blue, alpha)
	GetData(info)
	_settings[_field .. "R"], _settings[_field .. "G"], _settings[_field .. "B"], _settings[_field .. "A"] = red, green, blue, alpha
	RefreshSettings()
end

local function SetString(info, value)
	GetData(info)
	_settings[_category .. _field] = value
	RefreshSettings()
end

local function SetWithRefresh(info, value)
	Set(info, value)
	RefreshSettings()
end

--[[-----------------------------------------------------------------------------
String Options
-------------------------------------------------------------------------------]]
local string_settings = { Effect = true, Font = true, Size = true, VertAdj = true, Color = true, Override = true }

local function DisableStringEntry(info)
	GetData(info)
	if _category == "text" then
		return _settings.textDefaults
	elseif string_settings[_field] then
		return _settings[_category .. "Defaults"] or _settings[_category .. "UseText"]
	end
	return _settings[_category .. "Defaults"]
end

local function HideStringEntry(info)
	GetData(info)
	if _category == "text" then
		return _field ~= "Fixed"
	end
	return _field == "Fixed"
end

local stringArgs = {
	Defaults = {
		type = 'toggle',
		order = 1,
		name = "Defaults",
		desc = "Use defaults for these settings.",
		width = 'half',
		set = function(info, value)
			GetData(info)
			if value then
				MassClear(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Effect", "Fixed", "Font", "Hide", "Override", "Size", "UseText", "VertAdj")
			else
				_settings[_category .. "UseText"] = _settings[_category .. "UseText"]
				if not _settings[_category .. "UseText"] then
					MassLoad(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Effect", "Fixed", "Font", "Hide", "Override", "Size", "VertAdj")
				end
			end
			_settings[_category .. "Defaults"] = value
			RefreshSettings()
		end
	},
	Hide = {
		type = 'toggle',
		order = 2,
		name = "Hide",
		desc = "Determines if this item should be hidden.",
		width = 'half',
		disabled = DisableStringEntry
	},
	UseText = {
		type = 'toggle',
		order = 3,
		name = "Same as Text",
		desc = "Use the same settings as for text.",
		disabled = DisableStringEntry,
		hidden = HideStringEntry,
		set = function(info, value)
			GetData(info)
			if value then
				MassClear(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Effect", "Font", "Override", "Size", "VertAdj")
			else
				MassLoad(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Effect", "Font", "Override", "Size", "VertAdj")
			end
			_settings[_category .. "UseText"] = value
			RefreshSettings()
		end
	},
	Font = {
		type = 'select',
		order = 4,
		name = "Font",
		desc = "Choose a font to use.",
		dialogControl = 'LSM30_Font',
		values = AceGUIWidgetLSMlists.font,
		disabled = DisableStringEntry,
	},
	Effect = {
		type = 'select',
		order = 5,
		name = "Effect",
		desc = "Choose an effect to apply to the font.",
		values = { [0] = "None", [1] = "Outline", [2] = "Outline, Thick", [3] = "Shadow", [4] = "Shadow, Distinct" },
		disabled = DisableStringEntry,
	},
	Size = {
		type = 'range',
		order = 6,
		name = "Size",
		desc = "Set the size of the font.",
		disabled = DisableStringEntry,
		isPercent = true,
		min = 0.25, max = 1, step = 0.01
	},
	VertAdj = {
		type = 'range',
		order = 7,
		name = "Vertical Adjust",
		desc = "Align the font vertically.",
		disabled = DisableStringEntry,
		isPercent = true,
		min = -0.1, max = 0.1, step = 0.01
	},
	Color = {
		type = 'color',
		order = 8,
		name = "Color",
		desc = "Change the color of the font.",
		width = 'half',
		disabled = DisableStringEntry,
		get = function(info)
			GetData(info)
			return _settings[_category .. "ColorR"], _settings[_category .. "ColorG"], _settings[_category .. "ColorB"], _settings[_category .. "ColorA"]
		end,
		set = function(info, red, green, blue, alpha)
			GetData(info)
			_settings[_category .. "ColorR"], _settings[_category .. "ColorG"], _settings[_category .. "ColorB"], _settings[_category .. "ColorA"] = red, green, blue, alpha
			RefreshSettings()
		end
	},
	Override = {
		type = 'toggle',
		order = 9,
		name = "Override",
		desc = "Ignore native plugin font coloring.",
		width = 'half',
		disabled = DisableStringEntry
	},
	Fixed = {
		type = 'input',
		order = 10,
		name = "Fixed Width",
		desc = "Enter a value that best represents the width to be maintained.\n\nExample: 8888.88 DPS",
		disabled = DisableStringEntry,
		hidden = HideStringEntry,
		set = function(info, value)
			value = strtrim(tostring(value))
			SetString(info, value ~= "" and value or nil)
		end
	}
}

--[[-----------------------------------------------------------------------------
Plugin Options
-------------------------------------------------------------------------------]]
local function DisablePluginEntry(info)
	GetData(info)
	if _field == "tooltipScale" then
		return not _settings.tooltipAllowScaling or DisableIfDefault(info)
	elseif _root == "Base" then
		if _field == "lock" then
			return not AnyUnlocked("lockPlugin")
		end
		return not AnyLocked("lockPlugin")
	end
end

local function HidePluginEntry(info)
	GetData(info)
	if _root == "Plugins" then
		return _field == "lockPlugin" or _field == "lock" or _field == "unlock" or (_field == "delete" and plugins[PluginList.selection])
	elseif _root == "Panels" then
		return _field ~= "lockPlugin"
	else
		return not (_field == "lock" or _field == "unlock")
	end
end

local pluginOptions = {
	type = 'group',
	name = "Plugin Options",
	get = Get,
	set = SetWithRefresh,
	args = {
		desc = {
			type = 'description',
			order = 1,
			name = "These options allow you to change the display behavior of each plugin individually.\n",
			hidden = HidePluginEntry,
		},
--[[
		name = {
			type = 'description',
			order = 2,
			name = function()
				_temp = pluginNames[PluginList.selection]
				if plugins[_temp] then
					if plugins[_temp].data.label ~= _temp then
						return "   " .. _temp
					end
					return ""
				else
					return "   |cff808080< Offline >|r"
				end
			end,
			width = 'single',
			hidden = HidePluginEntry
		},
]]
		lockPlugin = {
			type = 'execute',
			order = 3,
			name = function(info)
				GetData(info)
				return _settings[_field] and "Plugins Locked" or "Plugins Unlocked"
			end,
			desc = function(info)
				GetData(info)
				return _settings[_field] and "Click to allow the plugins in this panel to be dragged by the mouse." or "Click to prevent the plugins in this panel from being dragged by the mouse."
			end,
			hidden = HidePluginEntry,
			func = function(info)
				GetData(info)
				_settings[_field] = not _settings[_field]
			end
		},
		lock = {
			type = 'execute',
			order = 4,
			name = "Lock",
			desc = "Click to prevent all plugins from being dragged by the mouse.",
			width = 'half',
			disabled = DisablePluginEntry,
			hidden = HidePluginEntry,
			func = function(info)
				settings.lockPlugin = true
				for id = 1, #panels do
					panels[id].settings.lockPlugin = nil
				end
			end
		},
		unlock = {
			type = 'execute',
			order = 5,
			name = "Unlock",
			desc = "Click to allow all plugins to be dragged by the mouse.",
			width = 'half',
			disabled = DisablePluginEntry,
			hidden = HidePluginEntry,
			func = function(info)
				settings.lockPlugin = nil
				for id = 1, #panels do
					panels[id].settings.lockPlugin = nil
				end
			end
		},
		general = {
			type = 'group',
			order = 1,
			name = "General",
			desc = "General settings.",
			hidden = HidePluginEntry,
			args = {
				enable = {
					type = 'toggle',
					order = 1,
					name = "Enable",
					desc = "Determines if this plugin should be shown or not.",
					width = 'half',
					set = function(info, value)
						Set(info, value)
						local name = PluginList.selection
						if value and not GetPluginLocation(name) then
							panels[1]:AddPlugin(name)
						elseif plugins[name] then
							plugins[name]:SetState()
						end
					end
				},
				delete = {
					type = 'execute',
					order = 2,
					name = "Delete",
					desc = "Delete these plugin settings.",
					width = 'half',
					hidden = HidePluginEntry,
					func = function()
						local name = PluginList.selection
						local panel, sectionType, index = GetPluginLocation(name)
						if panel then
							tremove(panel.settings[sectionType], index)
						end
						settings.plugins[name] = nil
						PluginList:Delete(name)
					end
				},
				panels = {
					type = 'select',
					order = 3,
					name = "Panel",
					desc = "Choose which panel should display this plugin.",
					values = panelList,
					get = function(info)
						_temp = GetPluginLocation(PluginList.selection)
						if not _temp then return end
						_temp = _temp.id
						for index = 1, #panelIDs do
							if panelIDs[index] == _temp then
								return index
							end
						end
					end,
					set = function(info, value)
						local panel, sectionType = GetPluginLocation(PluginList.selection)
						if panel ~= panels[panelIDs[value]] then
							panels[panelIDs[value]]:AddPlugin(PluginList.selection, sectionType)
						end
					end
				},
				section = {
					type = 'select',
					order = 4,
					name = "Section",
					desc = "Choose where in a panel to display this plugin.",
					values = section_types,
					get = function(info)
						return select(2, GetPluginLocation(PluginList.selection))
					end,
					set = function(info, value)
						local name = PluginList.selection
						local panel, sectionType = GetPluginLocation(name)
						if panel then
							if sectionType ~= value then
								panel:AddPlugin(name, value)
							end
						else
							panels[1]:AddPlugin(name, value)
						end
					end
				}
			}
		},
		icon = {
			type = 'group',
			name = "Icon",
			desc = "Icon related settings.",
			args = {
				iconDefaults = {
					type = 'toggle',
					order = 1,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					width = 'half',
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(_category, "Hide", "Size", "Zoom")
						else
							MassLoad(_category, "Hide", "Size", "Zoom")
						end
						_settings[_field] = value
						RefreshSettings()
					end
				},
				iconHide = {
					type = 'toggle',
					order = 2,
					name = "Hide",
					desc = "Determines if the icon should be hidden.",
					width = 'half',
					disabled = DisableIfDefault
				},
				iconSize = {
					type = 'range',
					order = 3,
					name = "Size",
					desc = "Set the size of the icon.",
					disabled = DisableIfDefault,
					isPercent = true,
					min = 0.25, max = 1, step = 0.01
				},
				iconZoom = {
					type = 'range',
					order = 4,
					name = "Zoom",
					desc = "Set the amount to change the icon's texture coordinates by.",
					disabled = DisableIfDefault,
					min = -0.5, max = 0.5, step = 0.01
				}
			}
		},
		label = {
			type = 'group',
			name = "Label",
			desc = "Label related settings.",
			get = GetString,
			set = SetString,
			args = stringArgs
		},
		suffix = {
			type = 'group',
			name = "Suffix",
			desc = "Suffix related settings.",
			get = GetString,
			set = SetString,
			args = stringArgs
		},
		text = {
			type = 'group',
			name = "Text",
			desc = "Text related settings.",
			get = GetString,
			set = SetString,
			args = stringArgs
		},
		tooltip = {
			type = 'group',
			name = "Tooltip",
			desc = "Tooltip related settings.",
			args = {
				tooltipDefaults = {
					type = 'toggle',
					order = 1,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					width = 'half',
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(_category, "AllowScaling", "HideCombat", "HideNoCombat", "Scale")
						else
							MassLoad(_category, "AllowScaling", "HideCombat", "HideNoCombat", "Scale")
						end
						_settings[_field] = value
					end
				},
				tooltipHideCombat = {
					type = 'toggle',
					order = 2,
					name = "Hide in Combat",
					desc = "Disable showing the tooltip while in combat.",
					disabled = DisableIfDefault
				},
				tooltipHideNoCombat = {
					type = 'toggle',
					order = 3,
					name = "Hide not in Combat",
					desc = "Disable showing the tooltip when not in combat.",
					disabled = DisableIfDefault
				},
				tooltipAllowScaling = {
					type = 'toggle',
					order = 4,
					name = "Allow Scaling",
					desc = "Enable tooltip scaling.\n\nWARNING: May conflict with some addons.",
					disabled = DisableIfDefault
				},
				tooltipScale = {
					type = 'range',
					order = 5,
					name = "Scale",
					desc = "Set the scale of the tooltip.",
					disabled = DisablePluginEntry,
					isPercent = true,
					min = 0.5, max = 1.5, step = 0.01
				}
			}
		}
	}
}

--[[-----------------------------------------------------------------------------
Panel Options
-------------------------------------------------------------------------------]]
local function DisablePanelEntry(info)
	GetData(info)
	if _root == "Base" then
		if _field == "lockPanel" then
			return not AnyUnlocked("lockPanel")
		end
		return not AnyLocked("lockPanel")
	end
end

local function HidePanelEntry(info)
	GetData(info)
	if _root == "Panels" then
		return _category ~= "general" and (_field == "lockPanel" or _field == "rightClickConfig" or _field == "template" or _field == "unlockPanel")
	end
	return not (_field == "lockPanel" or _field == "rightClickConfig" or _field == "template" or _field == "unlockPanel")
end

local panelOptions = {
	type = 'group',
	name = "Panel Options",
	get = Get,
	set = SetWithRefresh,
	args = {
		desc = {
			type = 'description',
			order = 1,
			name = "These options allow you to change the behavior of each panel individually.\n",
			hidden = HidePanelEntry
		},
		rightClickConfig = {
			type = 'toggle',
			order = 2,
			name = "Right Click Config",
			desc = "Determines if panels will open their configuration panel on a right click.",
			hidden = HidePanelEntry
		},
		template = {
			type = 'input',
			order = 3,
			name = "Label",
			desc = "Enter a template to use as a default panel label.\n\nExample: Panel $id of $#",
			hidden = HidePanelEntry,
			set = function(info, value)
				GetData(info)
				value, _temp = strtrim(RemoveColorCodes(value)), _settings[_field]
				if value ~= "" and value ~= DEFAULT_TEMPLATE then
					_settings[_field] = value
				else
					_settings[_field] = nil
					value = DEFAULT_TEMPLATE
				end
				if value ~= _temp then
					for id = 1, #panels do
						_temp = panels[id].settings
						if rawget(_temp, _field) == value then
							_temp[_field] = nil
						end
						PanelList:Add(id)
					end
					update_panel_list()
				end
			end
		},
		lockPanel = {
			type = 'execute',
			order = 4,
			name = "Lock",
			desc = "Click to prevent all panels from being dragged by the mouse.",
			width = 'half',
			disabled = DisablePanelEntry,
			hidden = HidePanelEntry,
			func = function(info)
				settings.lockPanel = true
				for id = 1, #panels do
					panels[id].settings.lockPanel = nil
				end
			end
		},
		unlockPanel = {
			type = 'execute',
			order = 5,
			name = "Unlock",
			desc = "Click to allow all panels to be dragged by the mouse.",
			width = 'half',
			disabled = DisablePanelEntry,
			hidden = HidePanelEntry,
			func = function(info)
				settings.lockPanel = nil
				for id = 1, #panels do
					panels[id].settings.lockPanel = nil
				end
			end
		},
		general = {
			type = 'group',
			order = 1,
			name = "General",
			desc = "General settings.",
			hidden = HidePanelEntry,
			args = {
				new = {
					type = 'execute',
					order = 1,
					name = "Create New Panel",
					desc = "Create a new panel.",
					hidden = HidePanelEntry,
					func = function()
						local panel = CreatePanel()
						update_panel_list()
						PanelList:Select(panel.id)
						panel:Refresh()
					end
				},
				enable = {
					type = 'toggle',
					order = 2,
					name = "Enable",
					desc = "Determines if this panel is active or not.",
					width = 'half',
					set = function(info, value)
						if value then
							panels[PanelList.selection]:Enable()
						else
							panels[PanelList.selection]:Disable()
						end
					end
				},
				delete = {
					type = 'execute',
					order = 3,
					name = "Delete",
					desc = "Delete this panel.",
					width = 'half',
					disabled = function()
						return #panels <= 1
					end,
					func = function()
						local id = PanelList.selection
						panels[id]:Delete()
						update_panel_list()
						PanelList:Select(min(id, #panels))
					end
				},
				template = {
					type = 'input',
					order = 4,
					name = "Label",
					desc = "Enter a template for this panel label.\n\nExample: Panel $id of $#",
					set = function(info, value)
						GetData(info)
						value, _temp = strtrim(RemoveColorCodes(value)), _settings[_field]
						if value ~= "" and value ~= settings[_field] then
							_settings[_field] = value
						else
							_settings[_field] = nil
						end
						if _settings[_field] ~= _temp then
							PanelList:Add(PanelList.selection)
							update_panel_list()
						end
					end
				},
				lockPanel = {
					type = 'execute',
					order = 5,
					name = function(info)
						GetData(info)
						return _settings[_field] and "Panel Locked" or "Panel Unlocked"
					end,
					desc = function(info)
						GetData(info)
						return _settings[_field] and "Click to allow this panel to be dragged by the mouse." or "Click to prevent this panel from being dragged by the mouse."
					end,
					hidden = HidePanelEntry,
					func = function(info)
						GetData(info)
						_settings[_field] = not _settings[_field]
					end
				}
			}
		},
		appearance = {
			type = 'group',
			name = "Appearance",
			desc = "Appearance related settings.",
			args= {
				appearanceDefaults = {
					type = 'toggle',
					order = 1,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(nil, "alpha", "height", "width")
						else
							MassLoad(nil, "alpha", "height", "width")
						end
						_settings[_field] = value
						RefreshSettings()
					end
				},
				height = {
					type = 'range',
					order = 2,
					name = "Height",
					desc = "Set the height.",
					disabled = DisableIfDefault,
					min = MIN_HEIGHT, max = MAX_HEIGHT, step = 0.01,
					get = function(info)
						GetData(info)
						return _settings[_field] * RESOLUTION_HEIGHT
					end,
					set = function(info, value)
						GetData(info)
						_settings[_field] = max(min(value, MAX_HEIGHT), MIN_HEIGHT) / RESOLUTION_HEIGHT
						RefreshSettings()
					end
				},
				width = {
					type = 'range',
					order = 3,
					name = "Width",
					desc = "Set the width.",
					disabled = DisableIfDefault,
					min = MIN_WIDTH, max = MAX_WIDTH, step = 0.01,
					get = function(info)
						GetData(info)
						return _settings[_field] * RESOLUTION_WIDTH
					end,
					set = function(info, value)
						GetData(info)
						_settings[_field] = max(min(value, MAX_WIDTH), MIN_WIDTH) / RESOLUTION_WIDTH
						RefreshSettings()
					end
				}
			}
		},
		bg = {
			type = 'group',
			name = "Background",
			desc = "Background related settings.",
			args = {
				bgDefaults = {
					type = 'toggle',
					order = 1,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					width = 'half',
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Inset", "Overlay", "OverlayColorA", "OverlayColorB", "OverlayColorG", "OverlayColorR", "OverlayFlip", "OverlayFlop", "Texture")
						else
							MassLoad(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Inset", "Overlay", "OverlayColorA", "OverlayColorB", "OverlayColorG", "OverlayColorR", "OverlayFlip", "OverlayFlop", "Texture")
						end
						_settings[_field] = value
						RefreshSettings()
					end
				},
				bgColor = {
					type = 'color',
					order = 2,
					name = "Color",
					desc = "Change the color and transparency of the background.",
					hasAlpha = true,
					width = 'half',
					disabled = DisableIfDefault,
					get = GetColor,
					set = SetColor
				},
				bgTexture = {
					type = 'select',
					order = 3,
					name = "Texture",
					desc = "Choose a texture to use for the background.",
					dialogControl = 'LSM30_Statusbar',
					values = AceGUIWidgetLSMlists.statusbar,
					disabled = DisableIfDefault
				},
				bgInset = {
					type = 'range',
					order = 4,
					name = "Inset",
					desc = "Set the distance inside the panel before the background begins.",
					disabled = DisableIfDefault,
					min = -1, max = 8, step = 1
				},
				bgOverlay = {
					type = 'select',
					order = 5,
					name = "Effect",
					desc = "Choose an effect to apply to the background texture.",
					dialogControl = 'LSM30_Overlay',
					values = AceGUIWidgetLSMlists.overlay,
					disabled = DisableIfDefault
				},
				bgOverlayFlip = {
					type = 'toggle',
					order = 6,
					name = "Flip",
					desc = "Reverse the effect horizontally.",
					width = 'half',
					disabled = DisableIfDefault
				},
				bgOverlayFlop = {
					type = 'toggle',
					order = 7,
					name = "Flop",
					desc = "Reverse the effect vertically.",
					width = 'half',
					disabled = DisableIfDefault
				},
				bgOverlayColor = {
					type = 'color',
					order = 8,
					name = "Tint",
					desc = "Change the color and transparency of the effect.",
					hasAlpha = true,
					width = 'half',
					disabled = DisableIfDefault,
					get = GetColor,
					set = SetColor
				}
			}
		},
		border = {
			type = 'group',
			name = "Border",
			desc = "Border related settings.",
			args = {
				borderDefaults = {
					type = 'toggle',
					order = 1,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					width = 'half',
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Size", "Texture")
						else
							MassLoad(_category, "ColorA", "ColorB", "ColorG", "ColorR", "Size", "Texture")
						end
						_settings[_field] = value
						RefreshSettings()
					end
				},
				borderColor = {
					type = 'color',
					order = 2,
					name = "Color",
					desc = "Change the color and transparency of the border.",
					width = 'half',
					hasAlpha = true,
					disabled = DisableIfDefault,
					get = GetColor,
					set = SetColor
				},
				borderTexture = {
					type = 'select',
					order = 3,
					name = "Texture",
					desc = "Choose a texture to use for the border.",
					dialogControl = 'LSM30_Border',
					values = AceGUIWidgetLSMlists.border,
					disabled = DisableIfDefault,
				},
				borderSize = {
					type = 'range',
					order = 4,
					name = "Size",
					desc = "Set how thick the border will be.",
					disabled = DisableIfDefault,
					min = 1, max = 16, step = 1
				}
			}
		},
		position = {
			type = 'group',
			name = "Position",
			desc = "Position related settings.",
			args= {
				reset = {
					type = 'execute',
					order = 1,
					name = "Reset Position",
					desc = "Reset position to the current anchor.",
					hidden = HidePanelEntry,
					func = function(info)
						panels[PanelList.selection]:SetOffset(0, 0)
					end
				},
				positionDefaults = {
					type = 'toggle',
					order = 2,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(nil, "anchor", "moveBlizzard", "screenClamp", "strata")
						else
							MassLoad(nil, "anchor", "moveBlizzard", "screenClamp", "strata")
						end
						_settings[_field] = value
						RefreshSettings()
					end
				},
				anchor = {
					type = 'select',
					order = 3,
					name = "Anchor",
					desc = "Choose the screen region for positioning to be relative to.",
					disabled = DisableIfDefault,
					values = anchor_points
				},
				moveBlizzard = {
					type = 'toggle',
					order = 4,
					name = "Move Blizzard Frames",
					desc = "Determines if the default Blizzard frames should be moved.",
					disabled = DisableIfDefault
				},
				screenClamp = {
					type = 'range',
					order = 5,
					name = "Screen Clamp",
					desc = "Set how much off-screen movement to allow.",
					disabled = DisableIfDefault,
					min = 0, max = 5, step = 1
				},
				strata = {
					type = 'select',
					order = 6,
					name = "Strata",
					desc = "Choose the strata layer.",
					disabled = DisableIfDefault,
					values = strata_layers
				}
			}
		},
		spacing = {
			type = 'group',
			name = "Spacing",
			desc = "Spacing related settings.",
			args= {
				spacingDefaults = {
					type = 'toggle',
					order = 1,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(_category, "Center", "Left", "LeftEdge", "Right", "RightEdge")
						else
							MassLoad(_category, "Center", "Left", "LeftEdge", "Right", "RightEdge")
						end
						_settings[_field] = value
						RefreshSettings()
					end
				},
				spacingLeftEdge = {
					type = 'range',
					order = 2,
					name = "Left Edge",
					desc = "Set how much distance to leave between the left panel edge and first left-side plugin.",
					disabled = DisableIfDefault,
					min = 0, max = 20, step = 1
				},
				spacingRightEdge = {
					type = 'range',
					order = 3,
					name = "Right Edge",
					desc = "Set how much distance to leave between the right panel edge and first right-side plugin.",
					disabled = DisableIfDefault,
					min = 0, max = 20, step = 1
				},
				spacingCenter = {
					type = 'range',
					order = 4,
					name = "Center Plugins",
					desc = "Set how much distance to leave between the center plugins.",
					disabled = DisableIfDefault,
					min = 0, max = 20, step = 1
				},
				spacingLeft = {
					type = 'range',
					order = 5,
					name = "Left Plugins",
					desc = "Set how much distance to place between left-side plugins.",
					disabled = DisableIfDefault,
					min = 0, max = 20, step = 1
				},
				spacingRight = {
					type = 'range',
					order = 6,
					name = "Right Plugins",
					desc = "Set how much distance to leave between right-side plugins.",
					disabled = DisableIfDefault,
					min = 0, max = 20, step = 1
				}
			}
		},
		alpha = {
			type = 'group',
			name = "Visibility",
			desc = "Visibility related settings.",
			args= {
				alphaDefaults = {
					type = 'toggle',
					order = 1,
					name = "Defaults",
					desc = "Use defaults for these settings.",
					set = function(info, value)
						GetData(info)
						if value then
							MassClear(_category, "Combat", "Faded", "Mouse", "Normal")
						else
							MassLoad(_category, "Combat", "Faded", "Mouse", "Normal")
						end
						_settings[_field] = value
						RefreshSettings()
					end
				},
				alphaNormal = {
					type = 'range',
					order = 2,
					name = "Alpha, Normal",
					desc = "Set the alpha to use on the panel.",
					disabled = DisableIfDefault,
					min = 0, max = 1, step = 0.05
				},
				alphaFaded = {
					type = 'range',
					order = 3,
					name = "Alpha, Faded",
					desc = "Set the alpha to use when the panel should be faded.",
					disabled = DisableIfDefault,
					min = 0, max = 1, step = 0.05
				},
				alphaCombat = {
					type = 'toggle',
					order = 4,
					name = "Combat Fading",
					desc = "Determines if the panel should be faded while in combat.",
					disabled = DisableIfDefault
				},
				alphaMouse = {
					type = 'toggle',
					order = 5,
					name = "Smart Mouse",
					desc = "Determines if the panel should change alpha based on whether or not the mouse is over it.",
					disabled = DisableIfDefault
				}
			}
		},
		plugins = {
			type = 'group',
			order = -1,
			name = "Plugin Defaults",
			desc = "Set the default behavior for plugins in this panel.",
			hidden = HidePanelEntry,
			args = pluginOptions.args
		}
	}
}

--[[-----------------------------------------------------------------------------
Base Options
-------------------------------------------------------------------------------]]
local baseOptions = {
	type = 'group',
	name = "Base Options",
	get = Get,
	set = Set,
	args = {
		desc = {
			type = 'description',
			order = 1,
			name = format("These options allow you to change the appearance and behavior of %s.\n", title)
		},
		general = {
			type = 'group',
			order = 1,
			name = "General",
			desc = "General settings.",
			args = {
				centerConfig = {
					type = 'toggle',
					order = 1,
					name = "Center Configuration",
					desc = "Keep the configuration window centered when the panel/plugin lists are showing."
				},
				report = {
					type = 'toggle',
					order = 2,
					name = "Report",
					desc = "Show warnings/errors generated by plugins that would otherwise be hidden."
				}
			}
		},
		panels = {
			type = 'group',
			name = "Panel Defaults",
			desc = "Set the default behavior for panels.",
			set = SetWithRefresh,
			args = panelOptions.args
		},
		plugins = {
			type = 'group',
			name = "Plugin Defaults",
			desc = "Set the default behavior for plugins.",
			set = SetWithRefresh,
			args = pluginOptions.args
		}
	}
}

--[[-----------------------------------------------------------------------------
Global to addon
-------------------------------------------------------------------------------]]
function UpdatePanelConfig()
	MAX_HEIGHT, MAX_WIDTH, MIN_HEIGHT, MIN_WIDTH = RESOLUTION_HEIGHT / 16, RESOLUTION_WIDTH, RESOLUTION_HEIGHT / 64, RESOLUTION_WIDTH / 64
	panelOptions.args.appearance.args.height.min = MIN_HEIGHT
	panelOptions.args.appearance.args.height.max = MAX_HEIGHT
	panelOptions.args.appearance.args.width.min = MIN_WIDTH
	panelOptions.args.appearance.args.width.max = MAX_WIDTH
	ConfigFrames[1]:Refresh()
end

function UpdatePluginConfig()
	ConfigFrames[2]:Refresh()
end

--[[-----------------------------------------------------------------------------
Initialize
-------------------------------------------------------------------------------]]
LibStub('AceConfigRegistry-3.0'):RegisterOptionsTable(title, {
	type = 'group',
	args = {
		Base = baseOptions,
		Panels = panelOptions,
		Plugins = pluginOptions,
		Profiles = LibStub('AceDBOptions-3.0'):GetOptionsTable(addon.db)
	}
})

ConfigFrames[0]:AssignOptions(title, "Base")
ConfigFrames[1]:AssignOptions(title, "Panels")
ConfigFrames[2]:AssignOptions(title, "Plugins")
ConfigFrames[3]:AssignOptions(title, "Profiles")

local function OnClose()
	local frame = GetMouseFocus()
	if type(frame) == 'table' and type(frame.id) == 'number' and type(frame.auxFrame) == 'table' and type(frame._SetWidth) == 'function' then
		frame:GetScript('OnLeave')(frame)
	end
	CONFIG_IS_OPEN = false
	AllPanels("UpdateVisibility")
end

local function OnOpen()
	local frame = GetMouseFocus()
	if type(frame) == 'table' and type(frame.id) == 'number' and type(frame.auxFrame) == 'table' and type(frame._SetWidth) == 'function' then
		frame:GetScript('OnEnter')(frame)
	end
	CONFIG_IS_OPEN = ConfigFrames[1]:IsShown() and PanelList.selection or true
	AllPanels("UpdateVisibility")
end

for index = 0, #ConfigFrames do
	ConfigFrames[index]:SetScript('OnClose', OnClose)
	ConfigFrames[index]:SetScript('OnOpen', OnOpen)
end

update_panel_list()
