--[[

	Pulse Info
	
	.categoryId			the category ID (string) that this alert falls into
	.name				the text associated with the pulse
	.texture			the texture to show when the pulse is shown
	.maxAlpha			the maximum alpha value the pulse should be shown at
	.fadeInTime			the time in seconds the pulse will take to fade in (i.e. go from 0 alpha to .maxAlpha)
	.holdTime			the time in seconds the pulse will remain at .maxAlpha before being faded out (nil = stay visible until manually faded)
	.fadeOutTime		the time in seconds the pulse will take to go from .maxAlpha back to 0
	.activateTime		the time at which the pulse will be shown
	
	
	
	Frame Configuration
	
	.name				the name of the frame (user-defined)
	.categories			a table of categoryId/boolean pairs of the alerts that the frame will show
	.position			a table containing Anchor, RelativeAnchor, X and Y values for the position of the frame (all based on UIParent)
	.scale				a floating-point number containing the scale of the frame (1 = 32 x 32)
	.filterCount		the number of items on the filter list (must not be nil)
	.filterList			a { name, boolean } table containing the 'names' of items on the filter list
	.filterType			an enum specifying the type of list ( 1 = black, 0 = white )
	
]]--

local CONST_CHAT_PREFIX_COLOR = { ["r"] = 1, ["g"] = 1, ["b"] = 0 }
local CONST_CHAT_PREFIX = "Ghost: Pulse - "

GhostPulse3 = {
	["locked"] = 1,
	["enabled"] = 1,
	["currentProfile"] = nil,
}

GhostPulse3DB = {
	["profiles"] = { },
	["currentProfile"] = "default",
}

-- locals
local LBF = nil
local GhostPulseLBFGroup = nil
local frm = CreateFrame("Frame")
local inCombat = false
local debugMode = false

-- functions
local function DecToHex(num)
    local b, k, rc, i, d = 16, "0123456789ABCDEF", "", 0


    while num > 0 do
        i = i + 1
        num, d = math.floor(num / b), num % b + 1
        rc = string.sub(k, d, d)..rc
    end
    
    return rc
end

local function FractionToHex(num)
    local rc = DecToHex(num * 255)


    if string.len(rc) == 1 then
        rc = "0"..rc
    elseif string.len(rc) == 0 then
    	rc = "00"
    end
    
    return rc
end

local function ColorToHex(col)
    return "FF"..FractionToHex(col.r)..FractionToHex(col.g)..FractionToHex(col.b)
end

local function ColoredText(text, color)
    return "|c"..ColorToHex(color)..text.."|r"
end

function GhostPulse3:TellUser(msg, r, g, b)
	local out = string.format("%s%s", ColoredText(CONST_CHAT_PREFIX, CONST_CHAT_PREFIX_COLOR), msg)
	
	
	DEFAULT_CHAT_FRAME:AddMessage(out, r, g, b)
end

local function SaveSkin(arg, SkinID, Gloss, Backdrop, Group, Button, Colors)
	GhostPulse3.currentProfile.SkinSettings = GhostPulse3.currentProfile.SkinSettings or { }
	
	GhostPulse3.currentProfile.SkinSettings.SkinID = SkinID
	GhostPulse3.currentProfile.SkinSettings.Gloss = Gloss
	GhostPulse3.currentProfile.SkinSettings.Backdrop = Backdrop
	GhostPulse3.currentProfile.SkinSettings.Colors = Colors
end

local function ApplyButtonFacadeSettings()
	if GhostPulse3.currentProfile.SkinSettings and GhostPulseLBFGroup then
		GhostPulseLBFGroup.Colors = GhostPulse3.currentProfile.SkinSettings.Colors or GhostPulseLBFGroup.Colors
		GhostPulseLBFGroup:Skin(GhostPulse3.currentProfile.SkinSettings.SkinID, GhostPulse3.currentProfile.SkinSettings.Gloss, GhostPulse3.currentProfile.SkinSettings.Backdrop)
	end
end

local function CreateZoomInPulseAnimation(frame)
    local group = frame:CreateAnimationGroup("Zoom")
    local inAnim = group:CreateAnimation("Alpha")
    local outAnim = group:CreateAnimation("Alpha")
    local zoomAnim = group:CreateAnimation("Scale")
    local setupAnim = group:CreateAnimation("Scale")
    
    
    setupAnim:SetScale(3, 3)
    setupAnim:SetDuration(0.01)
    setupAnim:SetOrder(1)
    
    inAnim:SetChange(1)
    inAnim:SetStartDelay(0)
    inAnim:SetOrder(2)
    inAnim:SetScript("OnUpdate", function(this)
		local alpha = 0 + (this.alpha - 0) * this:GetProgress()
		frame:SetAlpha(alpha)
    end)
    
    zoomAnim:SetScale(0.333333, 0.333333)
    zoomAnim:SetOrigin("CENTER", 0, 0)
    zoomAnim:SetOrder(2)
    
    outAnim:SetChange(-1)
    outAnim:SetOrder(2)

    return group, function(alpha, inTime, holdTime, outTime)
    	if alpha then
    		inAnim.alpha = alpha
    		inAnim:SetChange(alpha)
    	end
    	
    	if inTime and holdTime and outTime then
	    	local totalTime = inTime + holdTime + outTime
	    	
	    	
	    	inAnim:SetDuration(totalTime / 2)
	    	zoomAnim:SetDuration(totalTime)
	    	outAnim:SetStartDelay(totalTime / 2)
	    	outAnim:SetDuration(totalTime / 2)
	    end
    end
end

local function CreateZoomOutPulseAnimation(frame)
    local group = frame:CreateAnimationGroup("Zoom")
    local inAnim = group:CreateAnimation("Alpha")
    local outAnim = group:CreateAnimation("Alpha")
    local zoomAnim = group:CreateAnimation("Scale")
    
    
    inAnim:SetChange(1)
    inAnim:SetStartDelay(0)
    inAnim:SetOrder(1)
    inAnim:SetScript("OnUpdate", function(this)
		local alpha = 0 + (this.alpha - 0) * this:GetProgress()
		frame:SetAlpha(alpha)
    end)
    
    zoomAnim:SetScale(3, 3)
    zoomAnim:SetOrigin("CENTER", 0, 0)
    zoomAnim:SetOrder(1)
    
    outAnim:SetChange(-1)
    outAnim:SetOrder(1)

    return group, function(alpha, inTime, holdTime, outTime)
    	if alpha then
    		inAnim.alpha = alpha
    		inAnim:SetChange(alpha)
    	end
    	
    	if inTime and holdTime and outTime then
	    	local totalTime = inTime + holdTime + outTime
	    	
	    	
	    	inAnim:SetDuration(totalTime / 2)
	    	zoomAnim:SetDuration(totalTime)
	    	outAnim:SetStartDelay(totalTime / 2)
	    	outAnim:SetDuration(totalTime / 2)
	    end
    end
end

local function CreateStandardPulseAnimation(frame)
    local group = frame:CreateAnimationGroup("Zoom")
    local inAnim = group:CreateAnimation("Alpha")
    local outAnim = group:CreateAnimation("Alpha")
    
    
    inAnim:SetChange(1)
    inAnim:SetStartDelay(0)
    inAnim:SetOrder(1)
    inAnim:SetScript("OnUpdate", function(this)
		local alpha = 0 + (this.alpha - 0) * this:GetProgress()
		frame:SetAlpha(alpha)
    end)
    
    outAnim:SetChange(-1)
    outAnim:SetOrder(1)

    return group, function(alpha, inTime, holdTime, outTime)
    	if alpha then
    		inAnim.alpha = alpha
    		inAnim:SetChange(alpha)
    	end
    	
		if inTime then
	    	inAnim:SetDuration(inTime)
	    	
	    	if holdTime then
		    	outAnim:SetStartDelay(inTime + holdTime)
		    end
		end
		
		if outTime then
	    	outAnim:SetDuration(outTime)
	    end
    end
end

local function PlayPulseSound(categoryId, name, earlyWarning)
	local soundToPlay
	
	
	if not earlyWarning then
		soundToPlay = GhostPulse3.currentProfile.categories[categoryId] and GhostPulse3.currentProfile.categories[categoryId].sound or ""
	else
		soundToPlay = GhostPulse3.currentProfile.categories[categoryId] and GhostPulse3.currentProfile.categories[categoryId].earlyWarningSound or ""
	end
	
	if soundToPlay then
		PlaySound(soundToPlay)
		PlaySoundFile(soundToPlay)
	end
end

local function CreatePulseFrame(parent, name, profileName)
    local rc = CreateFrame("Button", "GhostPulseFrame_"..name.."_"..profileName, parent)
    
    
	local function TidyUp()
		rc.shown = false
		rc:Hide()
	end

    rc.name = name
    rc.locked = 1
    
    rc.image = rc:CreateTexture("GhostPulseTexture_"..name.."_"..profileName, "ARTWORK")
	rc.image:SetAllPoints(rc)
	
	rc:EnableMouse(false)
	
	rc.textFrame = rc:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	rc.textFrame:ClearAllPoints()
	rc.textFrame:SetPoint("CENTER", 0, 0)
	rc.textFrame:SetHeight(32)
	rc.textFrame:SetJustifyV("CENTER")
	rc.textFrame:SetJustifyH("CENTER")
	rc.textFrame:SetShadowColor(0, 0, 0)
	rc.textFrame:SetShadowOffset(0, 0)
	rc.textFrame:SetTextColor(1, 1, 1)
	rc.textFrame:SetText("")

	local anim, func
	
	rc.anim = { ["Zoom In"] = { }, ["Zoom Out"] = { }, ["Pulse"] = { }, }
	
    anim, func = CreateZoomOutPulseAnimation(rc, 1.5, 5)
    rc.anim["Zoom Out"].anim = anim
    rc.anim["Zoom Out"].SetAnimParams = func    
    rc.anim["Zoom Out"].anim:SetScript("OnFinished", TidyUp)
    
    anim, func = CreateZoomInPulseAnimation(rc, 1.5, 5)
    rc.anim["Zoom In"].anim = anim
    rc.anim["Zoom In"].SetAnimParams = func    
    rc.anim["Zoom In"].anim:SetScript("OnFinished", TidyUp)
    
    anim, func = CreateStandardPulseAnimation(rc, 1.5, 5)
    rc.anim["Pulse"].anim = anim
    rc.anim["Pulse"].SetAnimParams = func    
    rc.anim["Pulse"].anim:SetScript("OnFinished", TidyUp)
    
	rc:SetScript("OnUpdate", function(this)
		local curTime
		
		
		if this.nextFadeInTime ~= nil then
			curTime = GetTime()

			if this.nextFadeInTime <= curTime then
				this.textFrame:SetText(this.pulseName)
				
				if this.greyscale then
					this.image:SetDesaturated(1)
				else
					this.image:SetDesaturated(0)				
				end

				if this.anim[this.animType].anim:IsPlaying() then
					this.anim[this.animType].anim:Stop()
				end

--this:SetAlpha(this.maxAlpha)
				
				this.anim[this.animType].SetAnimParams(this.maxAlpha, this.fadeInTime, this.holdTime, this.fadeOutTime)
				this.anim[this.animType].anim:Play()
				
				this.nextFadeInTime = nil
				this.shown = true
				
				PlayPulseSound(this.categoryId, this.pulseName, this.greyscale)
			end
		end
	end)
	
	rc.HandleItem = function(this)
		local infoType, id, bookType = GetCursorInfo()
		local itemName
		
		
		if infoType == "spell" then
			itemName = GetSpellName(id, bookType)
		
		elseif infoType == "item" then
			itemName = GetItemInfo(id)
			
		end
		
		if this.name and itemName then
			if IsControlKeyDown() then
				GhostPulse3:RemoveFromFilterList(this.name, itemName)
			else
				GhostPulse3:AddToFilterList(this.name, itemName)
			end
		end
					
		ClearCursor()
	end
	
	rc:SetScript("OnReceiveDrag", rc.HandleItem)
	rc:SetScript("OnMouseUp", rc.HandleItem)

	rc:SetScript("OnDragStart", function(this)
		this:StartMoving()
	end)
	
	rc:SetScript("OnDragStop", function(this)
		local anchor, _, relativeAnchor, x, y

		
		this:StopMovingOrSizing()		
		anchor, _, relativeAnchor, x, y = this:GetPoint()
		
		GhostPulse3.currentProfile.frames[name].position.anchor = anchor
		GhostPulse3.currentProfile.frames[name].position.relativeanchor = relativeAnchor
		GhostPulse3.currentProfile.frames[name].position.x = x
		GhostPulse3.currentProfile.frames[name].position.y = y
	end)    		

	rc:SetScript("OnMouseWheel", function(this, value)
		local adjustment = value * 0.25
		
		
		if not IsShiftKeyDown() then
			local newScale = GhostPulse3.currentProfile.frames[name].scale + adjustment
			
			
			if newScale >= 0.25 and newScale <= 10 then
				GhostPulse3.currentProfile.frames[name].scale = newScale
				this:SetScale(newScale)

				GhostPulse3:TellUser(string.format("Scale of %s is now %0.2f", name, newScale))
			end
		else
			local newScale = GhostPulse3.currentProfile.frames[name].fontScale + adjustment
			
			
			if newScale >= 0.25 and newScale <= 3.0 then
				GhostPulse3.currentProfile.frames[name].fontScale = newScale
				this:SetFontScale(newScale)
				
				GhostPulse3:TellUser(string.format("Text scale of %s is now %0.2f", name, newScale))
			end
		end
	end)
	
	function rc:SetFontScale(value)
		rc.textFrame:SetFont("Fonts\\ARIALN.TTF", 3 * value, "OUTLINE")
	end
	
    function rc:SetSize(value)
        rc:SetWidth(value)
        rc:SetHeight(value)
        
        rc.textFrame:SetWidth(value - 4)
        rc.textFrame:SetHeight(value - 4)
    end
    
    function rc:SetLocked(value)
    	rc.locked = value
    	
    	if value ~= 1 and value ~= false and not rc.disabled then
    		rc:SetTexture("Interface\\Icons\\Spell_Nature_ElementalPrecision_1")
    		rc.textFrame:SetText(rc.name)
			rc:EnableMouse(true)
			rc:EnableMouseWheel(true)
			rc:SetMovable(true)
			rc:RegisterForDrag("LeftButton")
    		rc:SetAlpha(1)
    		rc:Show()
    	else
    		rc:EnableMouse(false)
			rc:EnableMouseWheel(false)
			rc:SetMovable(false)
    		rc:RegisterForDrag("LeftButton")
    		rc:Hide()
    	end
    end
    
	rc.SetTexture = function(this, ...)
		this.image:SetTexture(...)
	end
	
	rc.SetTexCoord = function(this, ...)
		this.image:SetTexCoord(...)
	end
	
    if GhostPulseLBFGroup then
	    GhostPulseLBFGroup:AddButton(rc, { ["Icon"] = rc.image } )
	end
    
    return rc
end

local function CheckLists(frame, name, id)
	local rc = true
	local strId = tostring(id)
	
	
	if frame and name then
		if GhostPulse3.currentProfile.frames[frame.name].filterCount > 0 then
			if GhostPulse3.currentProfile.frames[frame.name].filterType == 0 then
				-- whitelist
				if id ~= nil and GhostPulse3.currentProfile.frames[frame.name].filterList[strId] ~= nil then
					rc = GhostPulse3.currentProfile.frames[frame.name].filterList[strId]
				else
					rc = GhostPulse3.currentProfile.frames[frame.name].filterList[name]
				end
			else
				-- blacklist
				if id ~= nil and GhostPulse3.currentProfile.frames[frame.name].filterList[strId] ~= nil then
					rc = not GhostPulse3.currentProfile.frames[frame.name].filterList[strId]
				else
					rc = not GhostPulse3.currentProfile.frames[frame.name].filterList[name]
				end
			end
		end
	end
	
	return rc
end

local function AnyFrameShowingCategory(categoryId)
	local rc
	
	
	if categoryId == "startup" then
		rc = true

	else
		for name, info in pairs(GhostPulse3.currentProfile.frames) do
			local curFrame = getglobal("GhostPulseFrame_"..name.."_"..GhostPulse3DB.currentProfile)
			
			
			if curFrame then
				if info.categories[categoryId] then
					rc = true
					break
				end
			end
		end
	end
		
	return rc
end

local function GetGroupType()
	local rc = "solo"
	
	
	if GetNumRaidMembers() > 0 then
		rc = "raid"
		
	elseif GetNumPartyMembers() > 0 then
		rc = "party"
	end
	
	return rc
end

local function ShowPulse(pulseInfo, earlyWarning)
	local result
	local options = GhostPulse3.currentProfile.categories and GhostPulse3.currentProfile.categories[pulseInfo.categoryId] and GhostPulse3.currentProfile.categories[pulseInfo.categoryId].conditions
	
	
	-- check alert conditions
	if options ~= nil and options ~= "" then
		result = SecureCmdOptionParse(options)
	else
		result = "show"
	end
	
	if result == "show" then
		local gType = GetGroupType()
		local okay
		
		
		result = "hide"
		
		if gType == "solo" then
			okay = GhostPulse3.currentProfile.categories and GhostPulse3.currentProfile.categories[pulseInfo.categoryId] and GhostPulse3.currentProfile.categories[pulseInfo.categoryId].whileSolo

			if okay then
				result = "show"
			end
		
		elseif gType == "party" then
			okay = GhostPulse3.currentProfile.categories and GhostPulse3.currentProfile.categories[pulseInfo.categoryId] and GhostPulse3.currentProfile.categories[pulseInfo.categoryId].whileParty

			if okay then
				result = "show"
			end
		
		elseif gType == "raid" then
			okay = GhostPulse3.currentProfile.categories and GhostPulse3.currentProfile.categories[pulseInfo.categoryId] and GhostPulse3.currentProfile.categories[pulseInfo.categoryId].whileRaid

			if okay then
				result = "show"
			end
		end
	end
	
	if result == "show" then
		local whenDead = GhostPulse3.currentProfile.categories and GhostPulse3.currentProfile.categories[pulseInfo.categoryId] and GhostPulse3.currentProfile.categories[pulseInfo.categoryId].whileDead


		if not whenDead then
			if UnitIsDeadOrGhost("player") then
				result = "hide"
			end
		end
	end
	
	if result == "show" then
		if inCombat == true and not GhostPulse3.currentProfile.categories[pulseInfo.categoryId].inCombat then
			result = "hide"
		end

		if inCombat == false and not GhostPulse3.currentProfile.categories[pulseInfo.categoryId].outOfCombat then
			result = "hide"
		end
	end
	
	if pulseInfo.categoryId == "startup" then
		result = "show"
	end
	
	-- carry on
	if result == "show" then	
		for name, info in pairs(GhostPulse3.currentProfile.frames) do
			local curFrame = getglobal("GhostPulseFrame_"..name.."_"..GhostPulse3DB.currentProfile)
			
			
			if curFrame then
				if curFrame.locked and not curFrame.disabled then
					if info.categories[pulseInfo.categoryId] or pulseInfo.categoryId == "startup" then
						-- check the lists
						if CheckLists(curFrame, pulseInfo.name, pulseInfo.spellId) then
							local curTime = GetTime()
							
							
							if type(pulseInfo.texture) == "function" then
								pulseInfo.texture(curFrame.image)
							else
								curFrame:SetTexCoord(0, 1, 0, 1)
								curFrame:SetTexture(pulseInfo.texture)
							end
							
							if earlyWarning then
								curFrame.pulseName = ""
								curFrame.greyscale = true							
								curFrame.fadeInTime = pulseInfo.fadeInTime / 2
								curFrame.holdTime = pulseInfo.holdTime / 2		
								curFrame.fadeOutTime = pulseInfo.fadeOutTime / 2
							else
								if pulseInfo.categoryId == "startup" or GhostPulse3.currentProfile.categories[pulseInfo.categoryId].showText then
									curFrame.pulseName = pulseInfo.text
								else
									curFrame.pulseName = ""
								end
								
								curFrame.greyscale = nil
								curFrame.fadeInTime = pulseInfo.fadeInTime
								curFrame.holdTime = pulseInfo.holdTime		
								curFrame.fadeOutTime = pulseInfo.fadeOutTime									
							end

							curFrame.categoryId = pulseInfo.categoryId
							curFrame.maxAlpha = pulseInfo.maxAlpha
							curFrame.nextFadeOutTime = curTime + curFrame.holdTime
							curFrame.nextFadeInTime = curTime
							curFrame.color = pulseInfo.color
							curFrame.stacks = pulseInfo.stacks
							
							curFrame:Show()
						end
					end
				end
			end
		end
	end
end

local function OnUpdate(frame, elapsed)
	local curTime = GetTime()
	local somethingLeft
	
	
	for i, info in pairs(GhostPulse3.pulses) do
		if info.activateTime ~= nil then
			if GhostPulse3.currentProfile.categories[info.categoryId] and GhostPulse3.currentProfile.categories[info.categoryId].earlyWarning then
				if not info.earlyWarningIssued then
					if info.earlyWarningTime >= info.timeAdded then
						if info.earlyWarningTime <= curTime then
							local method = GhostPulse3.currentProfile.categories[info.categoryId].earlyWarningMethod or "Pulse"
							
							
							if method == "Both" or method == "Pulse" then
								ShowPulse(info, true)
							end
							
							if ( method == "Both" or ( method == "Text" or method == "MSBT" ) ) and CombatText_AddMessage ~= nil then
								CombatText_AddMessage(info.earlyText, CombatText_StandardScroll, 1, 1, 1, nil, false)
								PlaySound(GhostPulse3.currentProfile.categories[info.categoryId].earlyWarningSound or "")
							end
							
							if ( method == "Both" or ( method == "Text" or method == "MSBT" ) ) and CombatText_AddMessage == nil then
								if MikSBT and not MikSBT.IsModDisabled() then
									local tex
									
									
									if type(info.texture) == "string" then
										tex = info.texture
									end

									MikSBT.DisplayMessage(info.earlyText, MikSBT.DISPLAYTYPE_NOTIFICATION, false, 1.0 * 255, 1.0 * 255, 1.0 * 255, nil, nil, nil, tex)
									PlaySound(GhostPulse3.currentProfile.categories[info.categoryId].earlyWarningSound or "")
								end
							end
							
							info.earlyWarningIssued = true
							
							if GhostPulse3.categories[info.categoryId].callback then
								if type(GhostPulse3.categories[info.categoryId].callback) == "function" then
									GhostPulse3.categories[info.categoryId].callback("earlywarning", info.name)
								end
							end
						end
					end					
				end
			end
						
			if info.activateTime ~= nil and info.activateTime <= curTime then
				ShowPulse(info)
				info.activateTime = nil
				
				if GhostPulse3.currentProfile.categories[info.categoryId] and GhostPulse3.categories[info.categoryId].callback then
					if type(GhostPulse3.categories[info.categoryId].callback) == "function" then
						GhostPulse3.categories[info.categoryId].callback("pulse", info.name)
					end
				end					
			else
				somethingLeft = true
			end
		end
	end
	
	if not somethingLeft then
		GhostPulse3.pulses = { }
		frm:SetScript("OnUpdate", nil)
	end
end

local function RemoveFrame(this, name)
	if GhostPulse3.currentProfile.frames[name] then
		GhostPulse3.currentProfile.frames[name] = nil
		GhostPulse3.currentProfile.frameCount = currentProfile.frameCount - 1
	end
end

local function CreateAllFrames()
	LBF = LibStub and LibStub("LibButtonFacade", true)
	
	if LBF then
		if not GhostPulseLBFGroup then
			GhostPulseLBFGroup = LBF:Group("Ghost: Pulse")
			LBF:RegisterSkinCallback("Ghost: Pulse", SaveSkin, nil)	
		end
	end
			
	for name, info in pairs(GhostPulse3.currentProfile.frames) do
		local frameName = "GhostPulseFrame_"..name.."_"..GhostPulse3DB.currentProfile
		local curFrame = getglobal(frameName)
		
		
		if curFrame == nil then
			curFrame = CreatePulseFrame(UIParent, name, GhostPulse3DB.currentProfile)
			curFrame:SetPoint(info.position.anchor, "UIParent", info.position.relativeanchor, info.position.x, info.position.y)
			curFrame:SetSize(32)
			curFrame:SetScale(info.scale)
			curFrame:SetFontScale(info.fontScale)
			curFrame.animType = info.animType or "Pulse"
		end
	end
end

local function AddFrame(this, name)
	local rc
	
	
	if not GhostPulse3.currentProfile.frames[name] then
		local info = { }
	
	
		info.name = name
		info.scale = 5
		info.fontScale = 1
		info.categories = { }
		info.position = {
			["anchor"] = "CENTER",
			["relativeanchor"] = "CENTER",
			["x"] = 0,
			["y"] = 0,
		}
		info.animType = "Pulse"
		info.filterCount = 0
		info.filterType = 1 -- blacklist
		info.filterList = { }
		
		rc = info
		GhostPulse3.currentProfile.frames[name] = info
		GhostPulse3.currentProfile.frameCount = GhostPulse3.currentProfile.frameCount + 1
		
		CreateAllFrames()
		ApplyButtonFacadeSettings()
	end
	
	return rc
end

local function HandleAddonLoaded(addOnName)
	if addOnName == "GhostPulse3" then
		if GhostPulse3DB.profiles["default"] == nil then
			GhostPulse3.CreateProfile("default")
		end
		
		-- choose the profile for this character
		GhostPulse3DB.charProfiles = GhostPulse3DB.charProfiles or { }
		
		local charName = UnitName("player")
		local profName = GhostPulse3DB.charProfiles[charName]
		
		
		if profName == nil then
			profName = GhostPulse3DB.currentProfile
			GhostPulse3DB.charProfiles[charName] = profName
		end
			
		GhostPulse3.currentProfile = GhostPulse3DB.profiles[profName]

		-- remedy some defaults
		if GhostPulse3.currentProfile.activeIn == nil then
			GhostPulse3.currentProfile.activeIn = { ["none"] = 1, ["pvp"] = 1, ["arena"] = 1, ["party"] = 1, ["raid"] = 1, }
		end
		
		if GhostPulse3.currentProfile.frameCount == 0 then
			local f = GhostPulse3:AddFrame("Cooldowns")
			
			
			f.categories["spell"] = true
			f.categories["petspell"] = true
			f.categories["buffgain"] = true
			f.categories["bufffade"] = true
			f.categories["trinket"] = true
			f.categories["bag"] = true
			f.scale = 5
		end
		
		CreateAllFrames()
		ApplyButtonFacadeSettings()

		GhostPulse3:AddPulse("startup", "Initialize", nil, "Initialize", "Initialize", "Interface\\Icons\\Spell_Nature_ElementalPrecision_1", 0.5, 0.5, 0.5, GetTime() + 2)
	end
end

local eventHandlers = {
	["ADDON_LOADED"] = HandleAddonLoaded,
	["PLAYER_REGEN_ENABLE"] = function() inCombat = false end,
	["PLAYER_REGEN_DISABLE"] = function() inCombat = true end,	
}

local function OnEvent(frame, event, ...)
	local f = eventHandlers[event]
	
	
	if f then
		f(...)
	end
end

frm:SetScript("OnEvent", OnEvent)

for name, _ in pairs(eventHandlers) do
	frm:RegisterEvent(name)
end

local function AddPulse(engine, categoryId, name, spellId, text, earlyText, texture, fadeInTime, holdTime, fadeOutTime, activateTime)
	local slot
	local _, instType = IsInInstance()
	

	if GhostPulse3.enabled and GhostPulse3.currentProfile.activeIn[instType] and AnyFrameShowingCategory(categoryId) then
		-- check the lists first
		local okay = true

		
		if categoryId ~= "startup" then
			if GhostPulse3.currentProfile.categories[categoryId].filterCount > 0 then
				if GhostPulse3.currentProfile.categories[categoryId].filterType == 0 then
					okay = GhostPulse3.currentProfile.categories[categoryId].filterList[name]
				else
					okay = not GhostPulse3.currentProfile.categories[categoryId].filterList[name]
				end
			end
		end
						
		if okay then
			for i, info in pairs(engine.pulses) do
				if info.activateTime == nil then
					slot = i
					break
				end
			end
			
			if slot then
				engine.pulses[slot].timeAdded = GetTime()
				engine.pulses[slot].earlyWarningTime = activateTime - ((GhostPulse3.currentProfile.categories[categoryId] and GhostPulse3.currentProfile.categories[categoryId].earlyWarningTime) or 0.5)
				engine.pulses[slot].earlyWarningIssued = nil
				engine.pulses[slot].categoryId = categoryId
				engine.pulses[slot].name = name
				engine.pulses[slot].spellId = spellId
				engine.pulses[slot].text = text
				engine.pulses[slot].earlyText = earlyText
				engine.pulses[slot].texture = texture
				engine.pulses[slot].maxAlpha = 1
				engine.pulses[slot].fadeInTime = fadeInTime
				engine.pulses[slot].holdTime = holdTime
				engine.pulses[slot].fadeOutTime = fadeOutTime
				engine.pulses[slot].activateTime = activateTime
			else
				local info = { }
				
				info.timeAdded = GetTime()
				info.earlyWarningTime = activateTime - ((GhostPulse3.currentProfile.categories[categoryId] and GhostPulse3.currentProfile.categories[categoryId].earlyWarningTime) or 0.5)
				info.earlyWarningIssued = nil
				info.categoryId = categoryId
				info.name = name
				info.spellId = spellId
				info.text = text
				info.earlyText = earlyText
				info.texture = texture
				info.maxAlpha = 1
				info.fadeInTime = fadeInTime
				info.holdTime = holdTime
				info.fadeOutTime = fadeOutTime
				info.activateTime = activateTime
				
				table.insert(engine.pulses, info)
			end

			if GhostPulse3.enabled then
				frm:SetScript("OnUpdate", OnUpdate)
			end
		end
	end
end


local function RemovePulse(engine, categoryId, name)
	for i, info in pairs(engine.pulses) do
		if info.categoryId == categoryId then
			if info.name == name then
				info.activateTime = nil
			end
		end
	end
end

local function RegisterCategory(this, categoryId, categoryName, supportsEarlyWarnings, callback)
	GhostPulse3.categories = GhostPulse3.categories or { }
		
	GhostPulse3.categories[categoryId] = { }
	GhostPulse3.categories[categoryId].name = categoryName
	GhostPulse3.categories[categoryId].callback = callback
	GhostPulse3.categories[categoryId].supportsEarlyWarnings = supportsEarlyWarnings
	
	if GhostPulse3.currentProfile.categories[categoryId] == nil then
		GhostPulse3.currentProfile.categories[categoryId] = { }
		
		GhostPulse3.currentProfile.categories[categoryId].filterList = { }
		GhostPulse3.currentProfile.categories[categoryId].filterCount = 0
		GhostPulse3.currentProfile.categories[categoryId].filterType = 1 -- blacklist
		
		GhostPulse3.currentProfile.categories[categoryId].whileSolo = 1
		GhostPulse3.currentProfile.categories[categoryId].whileParty = 1
		GhostPulse3.currentProfile.categories[categoryId].whileRaid = 1
		
		GhostPulse3.currentProfile.categories[categoryId].earlyWarning = 1
		GhostPulse3.currentProfile.categories[categoryId].earlyWarningTime = 3
		
		GhostPulse3.currentProfile.categories[categoryId].inCombat = 1
		GhostPulse3.currentProfile.categories[categoryId].outOfCombat = 1
		
		GhostPulse3.currentProfile.categories[categoryId].earlyWarningMethod = "Pulse"
	end
end

-- profile management
local function CreateProfile(name)
	local rc = nil
	
	
	if GhostPulse3DB.profiles[name] == nil then
		GhostPulse3DB.profiles[name] = { }
		GhostPulse3DB.profiles[name].categories = { }
		GhostPulse3DB.profiles[name].frames = { }
		GhostPulse3DB.profiles[name].frameCount = 0
		rc = true
	end
	
	return rc
end

local function SwitchToProfile(name)
	local rc = nil
	
	if GhostPulse3DB.profiles[name] == nil then
		CreateProfile(name)
		
		if GhostPulse3.currentProfile.activeIn == nil then
			GhostPulse3.currentProfile.activeIn = { ["none"] = 1, ["pvp"] = 1, ["arena"] = 1, ["party"] = 1, ["raid"] = 1, }
		end
		
		if GhostPulse3.currentProfile.frameCount == 0 then
			local f = GhostPulse3:AddFrame("Cooldowns")
			
			
			f.categories["spell"] = true
			f.categories["petspell"] = true
			f.categories["buffgain"] = true
			f.categories["bufffade"] = true
			f.categories["trinket"] = true
			f.categories["bag"] = true
			f.scale = 5
		end
	end

	if GhostPulse3DB.profiles[name] then
		GhostPulse3DB.currentProfile = name
		GhostPulse3DB.charProfiles[UnitName("player")] = name		
		GhostPulse3.currentProfile = GhostPulse3DB.profiles[name]

		CreateAllFrames()
		ApplyButtonFacadeSettings()

		rc = true
	end
	
	return rc
end

local function RemoveProfile(name)
	local rc = nil
	
	
	if GhostPulse3DB.currentProfile ~= name then
		if GhostPulse3DB.profiles[name] then
			GhostPulse3DB.profiles[name] = nil
			rc = true
		end
	end
	
	return rc
end

-- master table
GhostPulse3.pulses = { }
GhostPulse3.AddPulse = AddPulse
GhostPulse3.RemovePulse = RemovePulse

GhostPulse3.AddFrame = AddFrame
GhostPulse3.RemoveFrame = RemoveFrame
GhostPulse3.RegisterCategory = RegisterCategory
GhostPulse3.debug = function(...)
	if debugMode then
		print(...)
	end
end

GhostPulse3.SetLocked = function(this, value)
	for name, _ in pairs(GhostPulse3.currentProfile.frames) do
		local frame = getglobal("GhostPulseFrame_"..name.."_"..GhostPulse3DB.currentProfile)
		
		
		if frame then
			frame:SetLocked(value)
		end
	end
	
	this.locked = value
end
GhostPulse3.CreateProfile = CreateProfile
GhostPulse3.SwitchToProfile = SwitchToProfile
GhostPulse3.RemoveProfile = RemoveProfile

-- slash command handling
local function Split(line)
	local first
	local second
	local posn = string.find(line," ")
	
	
	if posn ~= nil and posn > 0 then
		first = string.sub(line, 1, posn-1)
		second = string.sub(line, posn+1, string.len(line))
	else
		first = line
		second = nil
	end
	
	return first, second
end

local function ToggleLock()
	if GhostPulse3.locked == nil then
		GhostPulse3:SetLocked(1)
	else
		GhostPulse3:TellUser("Drag the frames to the position you want them.  Use the scroll wheel while over a frame to adjust its size, or use the shift key with the scroll wheel to adjust the text size.")
		GhostPulse3:SetLocked(nil)
	end
end

local commandDispatch = {
	["lock"] = ToggleLock,
	["enable"] = function()
		GhostPulse3.enabled = 1
		frm:SetScript("OnUpdate", OnUpdate)
		GhostPulse3:TellUser("Alerts will be displayed.")
	end,
	["disable"] = function()
		GhostPulse3.enabled = nil
		frm:SetScript("OnUpdate", nil)
		GhostPulse3.pulses = { }
		GhostPulse3:TellUser("Alerts will no longer be shown until you re-enable Ghost: Pulse, or reload your UI.")
	end,
	["debug"] = function()
		debugMode = not debugMode
		
		if debugMode then
			GhostPulse3:TellUser("Debugging output is now switched on.")
		else
			GhostPulse3:TellUser("Debugging output is now switched off.")
		end
	end,
}

local function CommandHandler(commandLine)
	local cmd, params = Split(commandLine)
	local func = commandDispatch[cmd]
	

	if func then
		func(params)
	else
		InterfaceOptionsFrame_OpenToCategory(GhostPulse3.optionsFrame)
	end
end

SlashCmdList["GHOSTPULSE"] = CommandHandler
SLASH_GHOSTPULSE1 = "/pulse"
SLASH_GHOSTPULSE2 = "/gp"

