--Local references are faster
local WatchFrame = WatchFrame
local math_max, math_min, math_ceil, table_insert, table_remove, table_wipe, string_format = math.max, math.min, math.ceil, table.insert, table.remove, table.wipe, string.format
local quixote = LibStub("LibQuixote-2.0")
local LSM = LibStub("LibSharedMedia-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Inquest", true)

function Inquest:InitializeTracker()
	--Increase the max number of watchable quests
	self.oldMaxWatchableQuests = MAX_WATCHABLE_QUESTS
	MAX_WATCHABLE_QUESTS = 25
	
	--Create a test line for getting width and height
	self.watchFrameTestLine = WatchFrame.lineCache:GetFrame()
	
	--Sets up ignoring Blizzard's settings for simple watch and ignore cursor
	self:RawHook("WatchFrame_ToggleSimpleWatch", true)
	self:RawHook("WatchFrame_ToggleIgnoreCursor", true)
	self:WatchFrame_ToggleSimpleWatch()
	self:WatchFrame_ToggleIgnoreCursor()
	
	--Make the watch frame movable from anywhere, not just the title bar
	WatchFrame:SetMovable(true)
	WatchFrame:RegisterForDrag("LeftButton")
	self:SecureHook("WatchFrame_Lock")
	self:SecureHook("WatchFrame_Unlock")
	self:SecureHook("WatchFrame_Collapse")
	self:SecureHook("WatchFrame_Expand")
	self:HookScript(WatchFrame, "OnDragStart", "WatchFrame_OnDragStart")
	self:HookScript(WatchFrame, "OnDragStop", "WatchFrame_OnDragStop")
	self:RawHookScript(WatchFrameTitleButton, "OnDragStart", "WatchFrame_OnDragStart")
	self:RawHookScript(WatchFrameTitleButton, "OnDragStop", "WatchFrame_OnDragStop")
	self:HookScript(WatchFrameLeftResizeThumb, "OnDragStop", function() self:SaveTrackerPosition() end)
	self:HookScript(WatchFrameRightResizeThumb, "OnDragStop", function() self:SaveTrackerPosition() end)
	
	--Increase the max size of the watch frame
	WatchFrame:SetMaxResize(600, UIParent:GetHeight())
	
	quixote.RegisterCallback(self, "Quest_Complete", "Quest_Complete")
	quixote.RegisterCallback(self, "Quest_Lost", "Quest_Lost")
	self:RegisterEvent("ACHIEVEMENT_EARNED")
	
	--Hooks for main processing functions
	self:RawHook("IsQuestWatched", true)
	self:RawHook("GetQuestIndexForWatch", true)
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	self:SecureHook("AddTrackedAchievement")
	self:SecureHook("RemoveTrackedAchievement")
	self:RawHook("WatchFrame_GetRemainingSpace", true)
	--self:RawHook("WatchFrame_GetHeightNeededForQuest", true)
	self:RawHook("WatchFrame_GetHeightNeededForAchievement", "GetHeightAndWidthNeededForAchievement", true)
	
	--Link title button dropdown functions to our functionality
	self:SecureHook("WatchFrame_OpenToObjectivesCategory")
	self:SecureHook("WatchFrame_SetBaseAlpha", function(alpha) self.db.profile.trackerBackdropAlpha = 1 - WatchFrame.baseAlpha end)
	
	--Allows for click bindings
	self:RawHook("WatchFrameLinkButtonTemplate_OnClick", true)
	
	--Temporary 3.2 fix for change in watched quest storage
	if(self.db.char.numQuestsWatched > 0 and type(self.db.char.watchedQuests[1]) == "table") then
		table_wipe(self.db.char.watchedQuests)
		self.db.char.numQuestsWatched = 0
	end
	
	--Wait until the first QUEST_LOG_UPDATE event to load watched quests from DB
	self:RegisterEvent("QUEST_LOG_UPDATE", "SyncWatches")
	
	self:LoadTrackerSettings()
	
	--Secure hook for auto show/hide
	self:SecureHook("WatchFrame_Update", "AutoHideTracker")
	
	--Hack to fix 3.1.2 issue with watch frame title button
	WatchFrameTitleButton:SetScript("OnMouseUp", nil)
	WatchFrameTitleButton:SetScript("OnClick", WatchFrameTitleButton_OnClick)

	WatchFrame_RemoveObjectiveHandler(WatchFrame_HandleDisplayQuestTimers)
	WatchFrame_RemoveObjectiveHandler(WatchFrame_HandleDisplayTrackedAchievements)
	WatchFrame_RemoveObjectiveHandler(WatchFrame_DisplayTrackedQuests)
	WatchFrame_AddObjectiveHandler(function(...) return self:ObjectiveDisplayHandler(...) end)
	
	WatchFrame_ClearDisplay()
end

--Prints the watch database, used for debugging
function Inquest:PrintWatchDB()
	for i=1, self.db.char.numQuestsWatched do
		self:Print("quest watch " .. i .. ": " .. self.db.char.watchedQuests[i])
	end
	for i=1, self.db.char.numAchievementsWatched do
		self:Print("achievement watch " .. i .. ": " .. self.db.char.watchedAchievements[i])
	end
end

local linkButtonIndex = 1
local linkButtons = {}
function Inquest:GetLinkButton()
	local button = linkButtons[linkButtonIndex]
	if(not button) then
		linkButtons[linkButtonIndex] = WatchFrame.buttonCache:GetFrame()
		button = linkButtons[linkButtonIndex]
	end
	
	if(not button.InquestHooked) then
		button.InquestHooked = true
		button:RegisterForDrag("LeftButton")
		self:HookScript(button, "OnDragStart", "WatchFrame_OnDragStart")
		self:HookScript(button, "OnDragStop", "WatchFrame_OnDragStop")
		self:HookScript(button, "OnEnter", "LinkButton_OnEnter")
		self:HookScript(button, "OnLeave", function() GameTooltip:Hide() end)
	end
	
	linkButtonIndex = linkButtonIndex + 1
	
	return button
end

function Inquest:ResetLinkButtons()
	linkButtonIndex = 1
end

function Inquest:ReleaseUnusedLinkButtons()
	local button
	for i=linkButtonIndex, #linkButtons do
		button = linkButtons[i]
		button.type = nil
		button.index = nil
		button:Hide()
		button.frameCache:ReleaseFrame(button)
		linkButtons[i] = nil
	end
end

local achievementLineIndex = 1
function Inquest:GetAchievementLine()
	local line = WATCHFRAME_ACHIEVEMENTLINES[achievementLineIndex]
	if(not line) then
		WATCHFRAME_ACHIEVEMENTLINES[achievementLineIndex] = WatchFrame.lineCache:GetFrame()
		line = WATCHFRAME_ACHIEVEMENTLINES[achievementLineIndex]
	end
	line:Reset()
	achievementLineIndex = achievementLineIndex + 1
	
	return line
end

function Inquest:ResetAchievementLines()
	achievementLineIndex = 1
end

function Inquest:ReleaseUnusedAchievementLines()
	local line
	for i=achievementLineIndex, #WATCHFRAME_ACHIEVEMENTLINES do
		line = WATCHFRAME_ACHIEVEMENTLINES[i]
		line:Hide()
		line.frameCache:ReleaseFrame(line)
		WATCHFRAME_ACHIEVEMENTLINES[i] = nil
	end
end

local questLineIndex = 1
function Inquest:GetQuestLine()
	local line = WATCHFRAME_QUESTLINES[questLineIndex]
	if(not line) then
		WATCHFRAME_QUESTLINES[questLineIndex] = WatchFrame.lineCache:GetFrame()
		line = WATCHFRAME_QUESTLINES[questLineIndex]
	end
	line:Reset()
	questLineIndex = questLineIndex + 1
	
	return line
end

function Inquest:ResetQuestLines()
	questLineIndex = 1
end

function Inquest:ReleaseUnusedQuestLines()
	local line
	for i=questLineIndex, #WATCHFRAME_QUESTLINES do
		line = WATCHFRAME_QUESTLINES[i]
		line:Hide()
		line.frameCache:ReleaseFrame(line)
		WATCHFRAME_QUESTLINES[i] = nil
	end
end

local timerLineIndex = 1
function Inquest:GetTimerLine()
	local line = WATCHFRAME_TIMERLINES[timerLineIndex]
	if(not line) then
		WATCHFRAME_TIMERLINES[timerLineIndex] = WatchFrame.lineCache:GetFrame()
		line = WATCHFRAME_TIMERLINES[timerLineIndex]
	end
	line:Reset()
	timerLineIndex = timerLineIndex + 1
	
	return line
end

function Inquest:ResetTimerLines()
	timerLineIndex = 1
end

function Inquest:ReleaseUnusedTimerLines()
	local line
	for i=timerLineIndex, #WATCHFRAME_TIMERLINES do
		line = WATCHFRAME_TIMERLINES[i]
		line:Hide()
		line:SetScript("OnEnter", nil)
		line:SetScript("OnLeave", nil)
		line:EnableMouse(false)
		line.frameCache:ReleaseFrame(line)
		WATCHFRAME_TIMERLINES[i] = nil
	end
end

--Returns the string that is stored in the DB for a given quest index
--[[function Inquest:GetSavedQuestString(questIndex)
	local title, level = GetQuestLogTitle(questIndex)
	if(title) then return title .. "/" .. level end
end]]

--Prevents watch frame simple mode from being toggled on
function Inquest:WatchFrame_ToggleSimpleWatch()
	if(WatchFrame.simpleMode) then
		ADVANCED_WATCH_FRAME = "1"
		WatchFrame.simpleMode = nil
		WatchFrame:SetUserPlaced(true)
		WatchFrame_Unlock(WatchFrame)
		WatchFrameTitleButton:Show()
		WatchFrameCollapseExpandButton:Show()
		WatchFrame_Update()
		WatchFrame:SetScript("OnUpdate", WatchFrame_OnUpdate)
		WatchFrame_ToggleIgnoreCursor()
	end
	if(ArenaEnemyFrames_UpdateWatchFrame) then
		ArenaEnemyFrames_UpdateWatchFrame()
	end
end

--Prevents ignore cursor from being toggled on
function Inquest:WatchFrame_ToggleIgnoreCursor()
	if(WATCHFRAME_IGNORECURSOR == "1") then
		WATCHFRAME_IGNORECURSOR = "0"
		WatchFrameTitleButton:Show()
		WatchFrameCollapseExpandButton:Enable()
	end
end

function Inquest:OnClick_Open(linkButton)
	CloseDropDownMenus()
	if(linkButton.type == "QUEST") then
		ExpandQuestHeader(GetQuestSortIndex(GetQuestIndexForWatch(linkButton.index)))
		QuestLog_OpenToQuest(GetQuestIndexForWatch(linkButton.index))
	elseif(linkButton.type == "ACHIEVEMENT") then
		if(not AchievementFrame) then 
			AchievementFrame_LoadUI() 
		end
		if(not AchievementFrame:IsShown()) then 
			AchievementFrame_ToggleAchievementFrame()
			AchievementFrame_SelectAchievement(linkButton.index)
		else
			if(AchievementFrameAchievements.selection ~= self.index) then
				AchievementFrame_SelectAchievement(linkButton.index)
			else
				AchievementFrame_ToggleAchievementFrame()
			end
		end
	end
end

function Inquest:OnClick_StopTracking(linkButton)
	if(linkButton.type == "QUEST") then
		RemoveQuestWatch(GetQuestIndexForWatch(linkButton.index))
		WatchFrame_Update()
		QuestLog_Update()
	elseif(linkButton.type == "ACHIEVEMENT") then
		RemoveTrackedAchievement(linkButton.index)
		WatchFrame_Update()
		if(AchievementFrame) then 
			AchievementFrameAchievements_ForceUpdate()
		end
	end
end

function Inquest:OnClick_ShareQuest(linkButton)
	if(linkButton.type == "QUEST" and GetQuestLogPushable(GetQuestIndexForWatch(linkButton.index))) then
		QuestLogPushQuest(GetQuestIndexForWatch(linkButton.index))
	end
end

function Inquest:OnClick_AbandonQuest(linkButton)
	if(linkButton.type == "QUEST") then
		local lastQuest = GetQuestLogSelection()
		
		SelectQuestLogEntry(GetQuestIndexForWatch(linkButton.index))
		SetAbandonQuest()
		local items = GetAbandonQuestItems()
		if(items) then
			StaticPopup_Hide("ABANDON_QUEST")
			StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
		else
			StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
			StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
		end
		SelectQuestLogEntry(lastQuest)
	end
end

function Inquest:OnClick_InsertChatLink(linkButton)
	if(ChatFrameEditBox:IsVisible()) then
		if(linkButton.type == "QUEST") then
			local questLink = GetQuestLink(GetQuestIndexForWatch(linkButton.index))
			if(questLink) then ChatEdit_InsertLink(questLink) end
		elseif(linkButton.type == "ACHIEVEMENT") then
			local achievementLink = GetAchievementLink(linkButton.index)
			if(achievementLink) then ChatEdit_InsertLink(achievementLink) end
		end
	end
end

function Inquest:OnClick_ShowBlizMenu(linkButton)
	local dropDown = WatchFrameDropDown
	CloseDropDownMenus()
	
	dropDown.type = linkButton.type
	dropDown.index = linkButton.index
	UIFrameFadeOut(WatchFrameLines, WATCHFRAME_FADETIME, WatchFrameLines:GetAlpha(), .5)
	WatchFrame.dropDownOpen = true
	WatchFrame.lastLinkButton = linkButton
	ToggleDropDownMenu(1, nil, dropDown, "cursor", 3, -3)
	linkButton:Disable()
	
	--Consider adding some sort of reference to WatchFrame.lastLinkButton here
end

--Handles link button clicks
function Inquest:WatchFrameLinkButtonTemplate_OnClick(linkButton, mouseButton, pushed)
	local clickBindingOptions = self.db.profile.clickBindings
	local command = nil
	if(mouseButton == "LeftButton") then
		if(IsShiftKeyDown()) then command = "shiftLeft"
		elseif(IsControlKeyDown()) then command = "ctrlLeft"
		elseif(IsAltKeyDown()) then command = "altLeft"
		else command = "left" end
	elseif(mouseButton == "RightButton") then
		if(IsShiftKeyDown()) then command = "shiftRight"
		elseif(IsControlKeyDown()) then command = "ctrlRight"
		elseif(IsAltKeyDown()) then command = "altRight"
		else command = "right" end
	end
	if(command) then
		local func = string.sub(clickBindingOptions[command], 3)
		if(func == "none") then return end
		
		self[func](self, linkButton)
	end
end

--Makes the title button dropdown "Settings" button open to Inquest settings
function Inquest:WatchFrame_OpenToObjectivesCategory()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.tracker)
end

--Handles quest completion
function Inquest:Quest_Complete(event, questTitle, uid)
	if(self.db.profile.questCompleteMessage) then
		UIErrorsFrame:AddMessage(string_format(L["%s Completed!"], questTitle), 1.0, 0.8, 0.0, 1.0, UIERRORS_HOLD_TIME)
	end
	
	if(self.db.profile.echoQuestCompleteToParty and GetNumPartyMembers() > 0 and GetNumRaidMembers() == 0) then
		SendChatMessage(string_format(L["%s Completed!"], questTitle), "PARTY")
	end
	
	if(self.db.profile.questCompleteSound) then
		PlaySoundFile(LSM:Fetch("sound", self.db.profile.questCompleteSoundFile))
	end
	
	if(self.db.profile.removeCompletedQuests) then
		local _, questIndex = quixote:GetQuestByUid(uid)
		RemoveQuestWatch(questIndex)
		WatchFrame_Update()
	end
end

--If we've lost a quest (from turnin or abandon), remove it from the DB
function Inquest:Quest_Lost(event, questTitle, questId)
	for i=1, self.db.char.numQuestsWatched do
		if(self.db.char.watchedQuests[i] == questId) then
			if(self.db.char.watchTimers[questId]) then
				self:CancelTimer(self.db.char.watchTimers[questId], true)
				self.db.char.watchTimers[questId] = nil
			end
			table_remove(self.db.char.watchedQuests, i)
			self.db.char.numQuestsWatched = self.db.char.numQuestsWatched - 1
			WatchFrame_Update()
			break
		end
	end
end

function Inquest:ACHIEVEMENT_EARNED(event, ...)
	self:RemoveTrackedAchievement(arg1)
end

function Inquest:SyncWatches()
	self:UnregisterEvent("QUEST_LOG_UPDATE")
	local questTitle, questIndex
	
	--Remove watches for quests that we don't have anymore
	local i = 1
	while(i <= self.db.char.numQuestsWatched) do
		questIndex = GetQuestIndexForWatch(i)
		if(questIndex) then 
			local questId = self:GetQuestId(questIndex)
			if(self.db.char.watchTimers[questId]) then
				table_remove(self.db.char.watchedQuests, i)
				self.db.char.watchTimers[questId] = nil
			end
			i = i + 1
		else
			table_remove(self.db.char.watchedQuests, i)
			self.db.char.numQuestsWatched = self.db.char.numQuestsWatched - 1
		end
	end
	
	--Add existing Blizzard watches to our DB
	local numQuestsWatched = GetNumQuestWatches()
	for i=1, numQuestsWatched do
		questIndex = GetQuestIndexForWatch(i)
		if(questIndex) then self:AddQuestWatch(questIndex) end
	end
	
	local trackedAchievements = { GetTrackedAchievements() }
	for i=1, #trackedAchievements do
		self:AddTrackedAchievement(trackedAchievements[i])
	end
	
	--Add our watches to Blizzard list, up to max watchable
	local numToWatch = math_min(self.db.char.numQuestsWatched, self.oldMaxWatchableQuests)
	for i=1, numToWatch do
		questIndex = GetQuestIndexForWatch(i)
		if(questIndex) then AddQuestWatch(questIndex) end
	end
	
	for i=1, self.db.char.numAchievementsWatched do
		AddTrackedAchievement(self.db.char.watchedAchievements[i])
	end
end

--Used to mass remove completed watches when that setting is toggled on
function Inquest:RemoveCompletedQuestWatches()
	local i = 1
	while i <= self.db.char.numQuestsWatched do
		local questIndex = GetQuestIndexForWatch(i)
		if(questIndex) then
			local title, _, _, _, _, _, isComplete = GetQuestLogTitle(questIndex)
			if(title and isComplete) then
				RemoveQuestWatch(questIndex)
			else
				i = i + 1
			end
		end
	end
	WatchFrame_Update()
end

function Inquest:WatchFrame_Lock()
	self.db.profile.trackerLocked = true
	WatchFrame:EnableMouse(false)
end

function Inquest:WatchFrame_Unlock()
	self.db.profile.trackerLocked = false
	WatchFrame:EnableMouse(true)
end

function Inquest:WatchFrame_Collapse(watchFrame)
	watchFrame:EnableMouse(false)
	self.db.char.watchFrameLastHeight = watchFrame:GetHeight()
	watchFrame:SetHeight(32)
	self:SetTrackerPosition(true)
end

function Inquest:WatchFrame_Expand(watchFrame)
	if(not self.db.profile.trackerLocked) then
		watchFrame:EnableMouse(true)
	end
	if(self.db.char.watchFrameLastHeight) then
		watchFrame:SetHeight(self.db.char.watchFrameLastHeight)
	end
	self:SetTrackerPosition()
end

function Inquest:WatchFrame_OnDragStart()
	if(not self.db.profile.trackerLocked) then
		WatchFrame.moving = true
		WatchFrame:StartMoving()
	end
end

function Inquest:WatchFrame_OnDragStop()
	if(WatchFrame.moving) then
		WatchFrame:StopMovingOrSizing()
		WatchFrame.moving = nil
		self:SaveTrackerPosition(WatchFrame)
		
		if(self.trackerUpdateWaiting) then 
			self.trackerUpdateWaiting = nil
			WatchFrame_Update() 
		end
	end
end

function Inquest:SaveTrackerPosition(watchFrame)
	watchFrame = watchFrame or WatchFrame
	self.db.profile.trackerInOriginalPosition = false
	self.db.profile.trackerTop = watchFrame:GetTop()
	self.db.profile.trackerBottom = watchFrame:GetBottom()
	self.db.profile.trackerLeft = watchFrame:GetLeft()
	self.db.profile.trackerRight = watchFrame:GetRight()
end

--Set current tracker state based on saved settings
function Inquest:LoadTrackerSettings()
	self.lastAcceptedQuest = nil
	
	if(WatchFrame.collapsed) then
		self:WatchFrame_Collapse(WatchFrame)
	else
		self:SetTrackerPosition()
	end
	
	--Set tracker locked state
	if(self.db.profile.trackerLocked) then 
		WatchFrame_Lock(WatchFrame)
	else
		WatchFrame_Unlock(WatchFrame)
	end
	
	--Make sure Blizzard auto tracking settings match up with ours
	local autoQuestWatch = (GetCVar("autoQuestWatch") == "1")
	local autoQuestProgress = (GetCVar("autoQuestProgress") == "1")
	if(self.db.profile.autoQuestWatch ~= autoQuestWatch) then
		local button = InterfaceOptionsObjectivesPanelAutoQuestTracking
		local setting = "0"
		if(self.db.profile.autoQuestWatch) then setting = "1" end
		SetCVar("autoQuestWatch", setting, button.event)
		AUTO_QUEST_WATCH = setting
		button:SetChecked(self.db.profile.autoQuestWatch)
		InterfaceOptionsPanel_CheckButton_Update(button)
	end
	if(self.db.profile.autoQuestProgress ~= autoQuestProgress) then
		local button = InterfaceOptionsObjectivesPanelAutoQuestProgress
		local setting = "0"
		if(self.db.profile.autoQuestProgress) then setting = "1" end
		SetCVar("autoQuestProgress", setting, button.event)
		AUTO_QUEST_PROGRESS = setting
		button:SetChecked(self.db.profile.autoQuestProgress)
		InterfaceOptionsPanel_CheckButton_Update(button)
	end
	
	if(self.db.char.watchFrameLastHeight) then
		WatchFrame:SetHeight(self.db.char.watchFrameLastHeight)
	end
	
	self:UpdateTrackerFontSpacing()
	WatchFrame_SetBaseAlpha(1 - self.db.profile.trackerBackdropAlpha)
	self:SetTrackerBackdropHidden()
	WATCHFRAME_QUEST_OFFSET = self.db.profile.trackerLineSpacing
end

function Inquest:SetTrackerPosition(collapse)
	if(WatchFrame.moving or WatchFrame.sizing) then return end
	
	WatchFrame:ClearAllPoints()
	if(collapse) then
		local yPoint, xPoint = self.db.profile.trackerCollapsePoint:match("(%a+)-(%a+)")
		local xOffset, yOffset = 0, 0
		if(yPoint == "TOP") then
			yOffset = self.db.profile.trackerTop
		else
			yOffset = self.db.profile.trackerBottom
		end
		
		if(xPoint == "RIGHT") then
			xOffset = self.db.profile.trackerRight
		else
			xOffset = self.db.profile.trackerLeft
		end
		
		WatchFrame:SetPoint(yPoint .. xPoint, UIParent, "BOTTOMLEFT", xOffset, yOffset)
	elseif(self.db.profile.trackerInOriginalPosition) then
		WatchFrame:SetPoint("TOPRIGHT", MinimapCluster, "BOTTOMRIGHT", 0, 10)
		self:SaveTrackerPosition(WatchFrame)
	else
		local anchorPoint, xOffset, yOffset = "", 0, 0
		if(self.db.profile.trackerVerticalGrowth == "1-up") then
			anchorPoint = "BOTTOM"
			yOffset = self.db.profile.trackerBottom
		else
			anchorPoint = "TOP"
			yOffset = self.db.profile.trackerTop
		end
		
		if(self.db.profile.trackerHorizontalGrowth == "1-left") then
			anchorPoint = anchorPoint .. "RIGHT"
			xOffset = self.db.profile.trackerRight
		else
			anchorPoint = anchorPoint .. "LEFT"
			xOffset = self.db.profile.trackerLeft
		end
		
		WatchFrame:SetPoint(anchorPoint, UIParent, "BOTTOMLEFT", xOffset, yOffset)
	end
end

function Inquest:UpdateTrackerFontSpacing(fontSize)
	fontSize = fontSize or self.db.profile.trackerFontSize
	WATCHFRAME_LINEHEIGHT = fontSize + 4
	WATCHFRAME_LINEHEIGHT_PROGRESSBAR = WATCHFRAME_LINEHEIGHT + 12
	WATCHFRAMELINES_FONTHEIGHT = fontSize
	WATCHFRAMELINES_FONTSPACING = (WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTHEIGHT) / 2
end

function Inquest:SetTrackerBackdropHidden(isHidden)
	isHidden = isHidden or self.db.profile.hideTrackerBackdrop
	if(isHidden) then
		WATCHFRAME_HOVERALPHA = 0
		WatchFrame_SetBaseAlpha(1)
		WatchFrameTitleButton:EnableMouse(false)
		WatchFrameCollapseExpandButton:Disable()
	else
		WATCHFRAME_HOVERALPHA = 1
		WatchFrameTitleButton:EnableMouse(true)
		WatchFrameCollapseExpandButton:Enable()
	end
end

function Inquest:SetTrackerVisibility(isVisible)
	if(isVisible) then
		WatchFrame:Show()
		if(not WatchFrame.collapsed) then WatchFrameLines:Show() end
	else
		WatchFrame:Hide()
		WatchFrameLines:Hide()
	end
end

function Inquest:AutoHideTracker()
	local hideWhen = self.db.profile.hideTrackerWhen
	local hideTracker = hideWhen.always
	if(not hideTracker) then
		local inInstance, instanceType = IsInInstance()
		if(inInstance) then
			hideTracker = (hideWhen.inDungeon and instanceType == "party") or 
				(hideWhen.inRaid and instanceType == "raid") or 
				(hideWhen.inBattleground and instanceType == "pvp") or 
				(hideWhen.inArena and instanceType == "arena")
		else
			hideTracker = (WatchFrame.nextOffset == WATCHFRAME_INITIAL_OFFSET and self.db.profile.hideTrackerWhen.empty)
		end
	end
	self:SetTrackerVisibility(not hideTracker)
end

function Inquest:GetQuestId(questIndex)
	return select(9, GetQuestLogTitle(questIndex))
end

function Inquest:IsQuestWatched(questIndex)
	--Call the Blizz function in case it has some kind of side effects
	local watchedByBlizz = self.hooks["IsQuestWatched"](questIndex)
	local questId = self:GetQuestId(questIndex)
	for i=1, self.db.char.numQuestsWatched do
		if(self.db.char.watchedQuests[i] == questId) then return true end
	end
	return false
end

function Inquest:GetQuestIndexForWatch(watchIndex)
	--Call the Blizz function in case it has some kind of side effects
	local blizzQuestIndex = self.hooks["GetQuestIndexForWatch"](watchIndex)
	if(watchIndex <= self.db.char.numQuestsWatched) then
		local numEntries = GetNumQuestLogEntries()
		for i=1, numEntries do
			if(self:GetQuestId(i) == self.db.char.watchedQuests[watchIndex]) then
				return i 
			end
		end
	end
	return nil
end

function Inquest:AddQuestWatch(questIndex, watchTime)
	local questId = self:GetQuestId(questIndex)
	for i=1, self.db.char.numQuestsWatched do
		if(self.db.char.watchedQuests[i] == questId) then
			if(self.db.char.watchTimers[questId]) then
				self:CancelTimer(self.db.char.watchTimers[questId], true)
				if(watchTime) then
					self.db.char.watchTimers[questId] = self:ScheduleTimer("QuestWatchExpired", watchTime, i)
				else
					self.db.char.watchTimers[questId] = nil
				end
			end
			return
		end
	end
	
	table_insert(self.db.char.watchedQuests, questId)
	if(watchTime) then
		self.db.char.watchTimers[questId] = self:ScheduleTimer("QuestWatchExpired", watchTime, self.db.char.numQuestsWatched + 1)
	end
	self.db.char.numQuestsWatched = self.db.char.numQuestsWatched + 1
end

function Inquest:RemoveQuestWatch(questIndex)
	local questId = self:GetQuestId(questIndex)
	for i=1, self.db.char.numQuestsWatched do
		if(self.db.char.watchedQuests[i] == questId) then
			if(self.db.char.watchTimers[questId]) then
				self:CancelTimer(self.db.char.watchTimers[questId], true)
				self.db.char.watchTimers[questId] = nil
			end
			table_remove(self.db.char.watchedQuests, i)
			self.db.char.numQuestsWatched = self.db.char.numQuestsWatched - 1
			break
		end
	end
end

--Handles quest timer expiration for timed watches
function Inquest:QuestWatchExpired(watchIndex)
	local questIndex = GetQuestIndexForWatch(watchIndex)
	if(questIndex) then
		RemoveQuestWatch(questIndex)
		WatchFrame_Update()
	end
end

function Inquest:AddTrackedAchievement(achIndex)
	for i=1, self.db.char.numAchievementsWatched do
		if(self.db.char.watchedAchievements[i] == achIndex) then return end
	end
	
	table_insert(self.db.char.watchedAchievements, achIndex)
	self.db.char.numAchievementsWatched = self.db.char.numAchievementsWatched + 1
end

function Inquest:RemoveTrackedAchievement(achIndex)
	for i=1, self.db.char.numAchievementsWatched do
		if(self.db.char.watchedAchievements[i] == achIndex) then
			table_remove(self.db.char.watchedAchievements, i)
			self.db.char.numAchievementsWatched = self.db.char.numAchievementsWatched - 1
			break
		end
	end
end

function Inquest:WatchFrame_GetRemainingSpace()
	--If auto resizing, max height is height of UIParent
	if(self.db.profile.autoResizeTracker) then
		if ( WATCHFRAME_NUM_TIMERS == 0 ) then
			return UIParent:GetHeight() - WATCHFRAMELINES_YOFFSET - WATCHFRAME_QUESTTIMER_HEIGHT - math.abs(WatchFrame.nextOffset);
		end
	 
		return (UIParent:GetHeight() - WATCHFRAMELINES_YOFFSET) - math.abs(WatchFrame.nextOffset);
	else
		return self.hooks["WatchFrame_GetRemainingSpace"]()
	end
end

function Inquest:SetTrackerHeightAndWidth(watchFrame)
	watchFrame = watchFrame or WatchFrame
	local totalHeightUsed = 0
	local maxWidth = 0
	
	local font = LSM:Fetch("font", self.db.profile.trackerFont)
	self.watchFrameTestLine.text:SetFont(font, self.db.profile.trackerFontSize)
	
	--Quest timers
	local numTimers = select("#", GetQuestTimers())
	if(numTimers > 0) then
		totalHeightUsed = totalHeightUsed + (WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING) * (numTimers + 1) + WATCHFRAME_TYPE_OFFSET
	end
	
	--Quests
	local iconHeightLeft, itemButton = 0, false
	local questWidth = 0
	local totalQuestHeight = 0
	for i=1, self.db.char.numQuestsWatched do
		questWidth = 0
		local questIndex = GetQuestIndexForWatch(i)
		if(questIndex) then
			totalQuestHeight = totalQuestHeight + WatchFrame_GetHeightNeededForQuest(questIndex) + 1
			
			--Title
			self.watchFrameTestLine.text:SetText(self:GetQuestLogTaggedTitle(questIndex))
			iconHeightLeft = 0
			itemButton = false
			if(GetQuestLogSpecialItemInfo(questIndex)) then
				itemButton = true
				iconHeightLeft = WATCHFRAME_QUEST_WITH_ITEM_HEIGHT - WATCHFRAMELINES_FONTHEIGHT - WATCHFRAMELINES_FONTSPACING
				questWidth = math_max(self.watchFrameTestLine.text:GetWidth() + WATCHFRAME_ITEM_WIDTH, questWidth)
			else
				questWidth = math_max(self.watchFrameTestLine.text:GetWidth(), questWidth)
			end
			
			--Objectives
			local numObjectives = GetNumQuestLeaderBoards(questIndex)
			for j=1, numObjectives do
				local text, _, finished = GetQuestLogLeaderBoard(j, questIndex)
				if(not finished or not self.db.profile.removeCompletedObjectives) then
					self.watchFrameTestLine.text:SetText(" - " .. text)
					if(itemButton and iconHeightLeft >= 0) then
						questWidth = math_max(self.watchFrameTestLine.text:GetWidth() + WATCHFRAME_ITEM_WIDTH, questWidth)
					else
						questWidth = math_max(self.watchFrameTestLine.text:GetWidth(), questWidth)
					end
					iconHeightLeft = iconHeightLeft - WATCHFRAMELINES_FONTHEIGHT - WATCHFRAMELINES_FONTSPACING
				end
			end
			
			if(itemButton) then
				maxWidth = math_max(questWidth + 8, maxWidth)
			else
				maxWidth = math_max(questWidth, maxWidth)
			end
		end
	end
	
	if(totalQuestHeight > 0) then
		totalHeightUsed = totalHeightUsed + WATCHFRAME_TYPE_OFFSET + totalQuestHeight
	end
	
	--Achievements with criteria
	local saveForLast = {}
	local totalAchievementHeight = 0
	local displayOnlyArena = (not watchFrame:IsUserPlaced()) and ArenaEnemyFrames and ArenaEnemyFrames:IsShown()
	for i=1, self.db.char.numAchievementsWatched do
		local achievementID = self.db.char.watchedAchievements[i]
		local achievementCategory = GetAchievementCategory(achievementID)
		if(not displayOnlyArena or achievementCategory == WATCHFRAME_ACHIEVEMENT_ARENA_CATEGORY) then
			local numCriteria = GetAchievementNumCriteria(achievementID)
			if(numCriteria == 0) then 
				table_insert(saveForLast, achievementID) 
			else
				local achievementHeight, achievementWidth = self:GetHeightAndWidthNeededForAchievement(achievementID)
				totalAchievementHeight = totalAchievementHeight + achievementHeight
				maxWidth = math_max(achievementWidth, maxWidth)
			end
		end
	end
	
	--Any other stuff that may have been added
	local lineFrame = WatchFrameLines
	local totalOffset = WATCHFRAME_INITIAL_OFFSET
	local maxHeight = UIParent:GetHeight()
	local maxFrameWidth = watchFrame:GetWidth() - WATCHFRAMELINES_XOFFSET
	for i=2, #WATCHFRAME_OBJECTIVEHANDLERS do
		local sectionHeightUsed, maxSectionWidth = WATCHFRAME_OBJECTIVEHANDLERS[i](lineFrame, totalOffset, maxHeight, maxFrameWidth)
		maxWidth = math_max(maxSectionWidth, maxWidth)
		
		if(sectionHeightUsed > 0) then
			totalHeightUsed = totalHeightUsed + WATCHFRAME_TYPE_OFFSET + sectionHeightUsed
		end
	end
	
	--Achievements without criteria
	for i=1, #saveForLast do
		local achievementCategory = GetAchievementCategory(saveForLast[i])
		if(not displayOnlyArena or achievementCategory == WATCHFRAME_ACHIEVEMENT_ARENA_CATEGORY) then 
			local achievementHeight, achievementWidth = self:GetHeightAndWidthNeededForAchievement(saveForLast[i], maxWidth)
			totalAchievementHeight = totalAchievementHeight + achievementHeight
			maxWidth = math_max(achievementWidth, maxWidth)
		end
	end
	
	if(totalAchievementHeight > 0) then
		totalHeightUsed = totalHeightUsed + WATCHFRAME_TYPE_OFFSET + totalAchievementHeight
	end
	
	if(self.db.profile.trackerHorizontalGrowth ~= "3-none") then
		watchFrame:SetWidth(math_max(WATCHFRAME_MINIMUMWIDTH, maxWidth + 26))
	end

	watchFrame:SetHeight(math_max(WATCHFRAME_MINIMUMHEIGHT, totalHeightUsed + 42))
	self:SetTrackerPosition()
	
	--Save new tracker position after resize
	self:SaveTrackerPosition(watchFrame)
end

function Inquest:ObjectiveDisplayHandler(lineFrame, initialOffset, maxHeight, frameWidth)
	if(self.db.profile.autoResizeTracker and not WatchFrame.collapsed) then
		WatchFrame.keepLeft = true
		self:SetTrackerHeightAndWidth(WatchFrame)
		maxHeight = WatchFrame:GetTop() - WatchFrame:GetBottom()
		frameWidth = WatchFrame:GetWidth() - WATCHFRAMELINES_XOFFSET
	end
	
	self:ResetLinkButtons()
	
	local heightUsed, sectionWidth, totalHeightUsed, maxWidth = 0, 0, 0, 0
	
	heightUsed, maxWidth = self:DisplayQuestTimers(lineFrame, initialOffset, maxHeight, frameWidth, GetQuestTimers())
	if(heightUsed > 0) then
		totalHeightUsed = totalHeightUsed + WATCHFRAME_TYPE_OFFSET + heightUsed
		initialOffset = initialOffset - WATCHFRAME_TYPE_OFFSET - heightUsed
	end
	
	heightUsed, sectionWidth = self:DisplayTrackedAchievements(lineFrame, initialOffset, maxHeight, frameWidth, GetTrackedAchievements())
	maxWidth = math_max(sectionWidth, maxWidth)
	if(heightUsed > 0) then
		totalHeightUsed = totalHeightUsed + WATCHFRAME_TYPE_OFFSET + heightUsed
		initialOffset = initialOffset - WATCHFRAME_TYPE_OFFSET - heightUsed
	end
	
	heightUsed, sectionWidth = self:DisplayTrackedQuests(lineFrame, initialOffset, maxHeight, frameWidth)
	maxWidth = math_max(sectionWidth, maxWidth)
	if(heightUsed > 0) then
		totalHeightUsed = totalHeightUsed + WATCHFRAME_TYPE_OFFSET + heightUsed
	end
	
	self:ReleaseUnusedLinkButtons()
	
	return totalHeightUsed, maxWidth
end

function Inquest:DisplayTrackedQuests(lineFrame, initialOffset, maxHeight, frameWidth)
	local maxWidth = 0
	local heightUsed = 0
	local iconHeightLeft = 0
	local watchItemIndex = 0
	local line, lastLine, questTitle
	local lastQuestHadItem
	
	self:ResetQuestLines()
	
	local font = LSM:Fetch("font", self.db.profile.trackerFont)

	for i=1, self.db.char.numQuestsWatched do
		local questWidth = 0
		local questIndex = GetQuestIndexForWatch(i)
		if(questIndex) then
			heightUsed = heightUsed + WatchFrame_GetHeightNeededForQuest(questIndex) + 1
			
			line = self:GetQuestLine()
			questTitle = line
			line.text:SetFont(font, self.db.profile.trackerFontSize)
			local title, completed = self:GetQuestLogTaggedTitle(questIndex)
			if(completed) then
				questTitle.text:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			else
				questTitle.text:SetTextColor(0.75, 0.61, 0)
			end
			line.text:SetText(title)
			
			local stringWidth = line.text:GetStringWidth()
			
			if(not lastLine) then
				line:SetPoint("TOPRIGHT", lineFrame, "TOPRIGHT", 0, initialOffset)
				line:SetPoint("TOPLEFT", lineFrame, "TOPLEFT", 0, initialOffset)
			else
				local yOffset = 0
				if(iconHeightLeft > 0) then yOffset = -iconHeightLeft end
				line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset - WATCHFRAME_QUEST_OFFSET)
				line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, yOffset - WATCHFRAME_QUEST_OFFSET)
			end
			line:Show()
			
			iconHeightLeft = 0
			local link, item, charges = GetQuestLogSpecialItemInfo(questIndex)
			local itemButton
			if(item and self.db.profile.showItemButtons) then
				watchItemIndex = watchItemIndex + 1
				itemButton = _G["WatchFrameItem" .. watchItemIndex]
				if(not itemButton) then
					WATCHFRAME_NUM_ITEMS = watchItemIndex
					itemButton = CreateFrame("Button", "WatchFrameItem" .. watchItemIndex, lineFrame, "WatchFrameItemButtonTemplate")
				end
				
				itemButton:ClearAllPoints()
				itemButton:SetID(questIndex)
				SetItemButtonTexture(itemButton, item)
				SetItemButtonCount(itemButton, charges)
				itemButton.charges = charges
				WatchFrameItem_UpdateCooldown(itemButton)
				itemButton.rangeTimer = -1
				
				if(self.db.profile.leftAlignItemButtons) then
					if(not lastLine) then
						itemButton:SetPoint("TOPLEFT", lineFrame, "TOPLEFT", 4, initialOffset)
						line:SetPoint("TOPRIGHT", lineFrame, "TOPRIGHT", 0, initialOffset)
					else
						local xOffset, yOffset = 0, 0
						if(lastQuestHadItem) then xOffset = -WATCHFRAME_ITEM_WIDTH - 8 end
						if(iconHeightLeft > 0) then yOffset = -iconHeightLeft end
						itemButton:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", xOffset + 4, yOffset - WATCHFRAME_QUEST_OFFSET)
						line:SetPoint("TOPRIGHT", lastLine, "TOPRIGHT", 0, yOffset - WATCHFRAME_QUEST_OFFSET)
					end
					line:SetPoint("TOPLEFT", itemButton, "TOPRIGHT", 8, 0)
				else
					line.text:SetPoint("RIGHT", itemButton, "LEFT", -4, 0)
				end
				
				iconHeightLeft = WATCHFRAME_QUEST_WITH_ITEM_HEIGHT - WATCHFRAMELINES_FONTHEIGHT - WATCHFRAMELINES_FONTSPACING
				itemButton.maxStringWidth = stringWidth
				questWidth = math_max(stringWidth + WATCHFRAME_ITEM_WIDTH, questWidth)
			else
				if(not lastLine) then
					line:SetPoint("TOPRIGHT", lineFrame, "TOPRIGHT", 0, initialOffset)
					line:SetPoint("TOPLEFT", lineFrame, "TOPLEFT", 0, initialOffset)
				else
					local xOffset, yOffset = 0, 0
					if(lastQuestHadItem) then xOffset = -WATCHFRAME_ITEM_WIDTH - 8 end
					if(iconHeightLeft > 0) then yOffset = -iconHeightLeft end
					line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset - WATCHFRAME_QUEST_OFFSET)
					line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", xOffset, yOffset - WATCHFRAME_QUEST_OFFSET)
				end
				
				iconHeightLeft = 0
				questWidth = math_max(stringWidth, questWidth)
			end
			line:Show()
			
			if(not WatchFrame.disableButtons) then
				linkButton = self:GetLinkButton()
				linkButton:SetPoint("TOPLEFT", line)
				linkButton:SetPoint("BOTTOMLEFT", line)
				linkButton:SetPoint("RIGHT", line.text)
				linkButton.type = "QUEST"
				linkButton.index = i
				linkButton:Show()
			end
			
			lastLine = line
			local numObjectives = GetNumQuestLeaderBoards(questIndex)
			for j=1, numObjectives do
				local text, _, finished = GetQuestLogLeaderBoard(j, questIndex)
				if(not finished or not self.db.profile.removeCompletedObjectives) then
					line = self:GetQuestLine()
					line.text:SetFont(font, self.db.profile.trackerFontSize)
					line.text:SetText(" - " .. text)
					if(finished) then
						line.text:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
					else
						line.text:SetTextColor(0.8, 0.8, 0.8)
					end
					
					line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, WATCHFRAMELINES_FONTSPACING)
					line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, WATCHFRAMELINES_FONTSPACING)
					stringWidth = line.text:GetStringWidth()
					if(itemButton) then
						if(self.db.profile.leftAlignItemButtons) then
							line:SetPoint("LEFT", itemButton, "RIGHT", 8, 0)
							questWidth = math_max(stringWidth + WATCHFRAME_ITEM_WIDTH, questWidth)
						elseif(iconHeightLeft >= 0) then
							line.text:SetPoint("RIGHT", itemButton, "LEFT", -4, 0)
							itemButton.maxStringWidth = math_max(stringWidth, itemButton.maxStringWidth)
							questWidth = math_max(stringWidth + WATCHFRAME_ITEM_WIDTH, questWidth)
						end
					else
						questWidth = math_max(stringWidth, questWidth)
					end
					line:Show()

					lastLine = line
					iconHeightLeft = iconHeightLeft - WATCHFRAMELINES_FONTHEIGHT - WATCHFRAMELINES_FONTSPACING
				end
			end
			
			if(itemButton) then
				if(self.db.profile.leftAlignItemButtons) then
					lastQuestHadItem = true
				else
					local xOffset = itemButton.maxStringWidth + 8
					if(not self.db.profile.autoResizeTracker or self.db.profile.trackerHorizontalGrowth == "3-none") then
						xOffset = math_min(frameWidth - WATCHFRAME_ITEM_WIDTH, xOffset)
					end
					itemButton:SetPoint("TOPLEFT", questTitle, "TOPLEFT", xOffset, -WATCHFRAMELINES_FONTSPACING);
				end
				itemButton:Show()
				maxWidth = math_max(questWidth + 8, maxWidth)
			else
				lastQuestHadItem = nil
				maxWidth = math_max(questWidth, maxWidth)
			end
		end
	end
	
	for i=watchItemIndex+1, WATCHFRAME_NUM_ITEMS do
		_G["WatchFrameItem" .. i]:Hide()
	end
	
	self:ReleaseUnusedQuestLines()
	
	return heightUsed, maxWidth
end

function Inquest:GetHeightAndWidthNeededForAchievement(achievementID, currentMaxWidth)
	local _, achievementName, _, completed, _, _, _, description = GetAchievementInfo(achievementID)
	if(completed) then return 0, 0 end
	
	--Title height and width
	local heightUsed = WATCHFRAME_LINEHEIGHT
	if(GetNumTrackedAchievements() > 1) then
		heightUsed = heightUsed + WATCHFRAME_QUEST_OFFSET
	end
	self.watchFrameTestLine.text:SetText(achievementName)
	local maxWidth = self.watchFrameTestLine.text:GetStringWidth()
	if(currentMaxWidth) then
		maxWidth = math_max(currentMaxWidth, maxWidth)
	end
	
	local lineHeight = WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING
	local numCriteria = GetAchievementNumCriteria(achievementID)
	if(numCriteria == 0) then
		local frameWidth = WatchFrame:GetWidth() - WATCHFRAMELINES_XOFFSET		
		self.watchFrameTestLine.text:SetText(" - ")
		local dashWidth = self.watchFrameTestLine.text:GetStringWidth()
		
		self.watchFrameTestLine.text:SetText(description)
		local stringWidth = self.watchFrameTestLine.text:GetStringWidth()
		local desiredWidth = math_ceil(stringWidth + dashWidth)
		if(desiredWidth > maxWidth and maxWidth < MAX_LONG_CRITERIA_WIDTH) then 
			maxWidth = math_min(desiredWidth, MAX_LONG_CRITERIA_WIDTH)
		end
		
		local linesNeeded = math_ceil(stringWidth / (maxWidth - dashWidth))
		heightUsed = heightUsed + (WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING) + (WATCHFRAMELINES_FONTHEIGHT * (linesNeeded - 1))
		
		for criteriaID, timedCriteria in next, WATCHFRAME_TIMEDCRITERIA do
			if(timedCriteria.achievementID == achievementID) then
				heightUsed = heightUsed + lineHeight
			end
		end
	else
		local criteriaDisplayed = 0
		for i=1, numCriteria do
			local criteriaString, criteriaType, criteriaCompleted, quantity, totalQuantity, name, flags, assetID, quantityString, criteriaID = GetAchievementCriteriaInfo(achievementID, i)
			if(criteriaCompleted or criteriaDisplayed > WATCHFRAME_CRITERIA_PER_ACHIEVEMENT) then 
				--Don't do anything
			elseif(criteriaDisplayed == WATCHFRAME_CRITERIA_PER_ACHIEVEMENT) then
				heightUsed = heightUsed + lineHeight
				self.watchFrameTestLine.text:SetText("...")
				maxWidth = math_max(self.watchFrameTestLine.text:GetStringWidth(), maxWidth)
				criteriaDisplayed = criteriaDisplayed + 1
			else
				if(WATCHFRAME_TIMEDCRITERIA[criteriaID]) then
					heightUsed = heightUsed + lineHeight
					criteriaDIsplayed = criteriaDisplayed + 1
				end
				if(bit.band(flags, ACHIEVEMENT_CRITERIA_PROGRESS_BAR) == ACHIEVEMENT_CRITERIA_PROGRESS_BAR) then
					heightUsed = heightUsed + WATCHFRAME_LINEHEIGHT_PROGRESSBAR
					maxWidth = math_max(200, maxWidth)
					criteriaDIsplayed = criteriaDisplayed + 2
				else
					heightUsed = heightUsed + lineHeight
					self.watchFrameTestLine.text:SetText(" - " .. criteriaString)
					maxWidth = math_max(self.watchFrameTestLine.text:GetStringWidth(), maxWidth)
					criteriaDisplayed = criteriaDisplayed + 1
				end
			end
		end
	end
	
	return heightUsed, maxWidth
end

function Inquest:DisplayTrackedAchievements(lineFrame, initialOffset, maxHeight, frameWidth, ...)
	local maxWidth = 0
	local heightUsed = 0
	local nextXOffset = 0
	local line, lastLine, achievementTitle
	
	self:ResetAchievementLines()
	
	local font = LSM:Fetch("font", self.db.profile.trackerFont)
	
	local numTrackedAchievements = select("#", ...)
	local displayOnlyArena = (not WatchFrame:IsUserPlaced()) and ArenaEnemyFrames and ArenaEnemyFrames:IsShown()
	for i=1, numTrackedAchievements do
		local achievementID = select(i, ...)
		local achievementCategory = GetAchievementCategory(achievementID)
		if(not displayOnlyArena or achievementCategory == WATCHFRAME_ACHIEVEMENT_ARENA_CATEGORY) then
			local _, achievementName, _, completed, _, _, _, description, _, icon = GetAchievementInfo(achievementID)
			
			heightUsed = heightUsed + WatchFrame_GetHeightNeededForAchievement(achievementID)
			
			line = self:GetAchievementLine()
			achievementTitle = line
			line.text:SetFont(font, self.db.profile.trackerFontSize)
			line.text:SetText(achievementName)
			if(completed) then
				line.text:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			else
				line.text:SetTextColor(0.75, 0.61, 0)
			end
			line:Show()
			local lineWidth = line.text:GetStringWidth() + WATCHFRAME_ICONXOFFSET
			maxWidth = math_max(maxWidth, lineWidth)
			if(lastLine) then
				local yOffset = 0
				if(lastLine.statusBar:IsShown()) then yOffset = -5 end
				
				line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset - WATCHFRAME_QUEST_OFFSET)
				line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", nextXOffset, yOffset - WATCHFRAME_QUEST_OFFSET)
			else
				line:SetPoint("TOPRIGHT", lineFrame, "TOPRIGHT", 0, initialOffset)
				line:SetPoint("TOPLEFT", lineFrame, "TOPLEFT", 0, initialOffset)
			end
			if(not WatchFrame.disableButtons) then
				linkButton = self:GetLinkButton()
				linkButton:SetPoint("TOPLEFT", line.text)
				linkButton:SetPoint("BOTTOMLEFT", line.text)
				linkButton:SetWidth(lineWidth + WATCHFRAME_ICONXOFFSET)
				linkButton.type = "ACHIEVEMENT"
				linkButton.index = achievementID
				linkButton:Show()
			end
			nextXOffset = 0
			lastLine = line
			
			local numCriteria = GetAchievementNumCriteria(achievementID)
			if(numCriteria > 0) then
				local criteriaDisplayed = 0
				for j=1, numCriteria do
					local criteriaString, criteriaType, criteriaCompleted, quantity, totalQuantity, name, flags, assetID, quantityString, criteriaID = GetAchievementCriteriaInfo(achievementID, j)
					if(criteriaCompleted or criteriaDisplayed > WATCHFRAME_CRITERIA_PER_ACHIEVEMENT) then
						--Don't do anything
					elseif(criteriaDisplayed == WATCHFRAME_CRITERIA_PER_ACHIEVEMENT) then
						line = self:GetAchievementLine()
						line.text:SetFont(font, self.db.profile.trackerFontSize)
						line.text:SetText(" - ")
						local dashWidth = line.text:GetStringWidth()
						nextXOffset = nextXOffset + dashWidth
						
						line.text:SetText("...")
						line.text:SetTextColor(0.8, 0.8, 0.8)
						line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", nextXOffset, WATCHFRAMELINES_FONTSPACING)
						line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, WATCHFRAMELINES_FONTSPACING)
						line:Show()
						
						maxWidth = math_max(maxWidth, line.text:GetStringWidth())
						criteriaDisplayed = criteriaDisplayed + 1
						lastLine = line
						nextXOffset = -dashWidth
					else
						if(WATCHFRAME_TIMEDCRITERIA[criteriaID]) then
							local timedCriteria = WATCHFRAME_TIMEDCRITERIA[criteriaID]
							line = self:GetAchievementLine()
							line.criteriaID = criteriaID
							line.duration = timedCriteria.duration
							line.startTime = timedCriteria.startTime
							line.text:SetFont(font, self.db.profile.trackerFontSize)
							
							local yOffset = WATCHFRAMELINES_FONTSPACING
							if(lastLine.statusBar:IsShown()) then yOffset = yOffset - 5 end
							
							line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset)
							line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", nextXOffset, yOffset)
							line:Show()
							criteriaDisplayed = criteriaDisplayed + 1
							lastLine = line
							nextXOffset = 0
							WatchFrameLines_AddUpdateFunction(WatchFrame_UpdatedTimedAchievements)
						end
						if(bit.band(flags, ACHIEVEMENT_CRITERIA_PROGRESS_BAR) == ACHIEVEMENT_CRITERIA_PROGRESS_BAR) then
							line = self:GetAchievementLine()
							line.statusBar:Show()
							line.statusBar:GetStatusBarTexture():SetVertexColor(0, 0.6, 0, 1)
							line.statusBar:SetMinMaxValues(0, totalQuantity)
							line.statusBar:SetValue(quantity)
							line.statusBar.text:SetFont(font, self.db.profile.trackerFontSize)
							line.statusBar.text:SetText(quantityString)
							line.statusBar.text:SetWidth(frameWidth)
							
							local yOffset = -5
							if(lastLine.statusBar:IsShown()) then yOffset = -10 end
							
							line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset)
							line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", nextXOffset, yOffset)
							line:Show()
							maxWidth = math_max(maxWidth, 200)
							criteriaDisplayed = criteriaDisplayed + 2
							nextXOffset = 0
							lastLine = line
						else
							line = self:GetAchievementLine()
							line.text:SetFont(font, self.db.profile.trackerFontSize)
							line.text:SetText(" - " .. criteriaString)
							line.text:SetTextColor(0.8, 0.8, 0.8)
							local yOffset = WATCHFRAMELINES_FONTSPACING
							if(lastLine.statusBar:IsShown()) then yOffset = yOffset - 5 end
							
							line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset)
							line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", nextXOffset, yOffset)
							line:Show()
							maxWidth = math_max(maxWidth, line.text:GetStringWidth())
							criteriaDisplayed = criteriaDisplayed + 1
							nextXOffset = 0
							lastLine = line
						end
					end
				end
			else
				local dash = self:GetAchievementLine()
				dash.text:SetFont(font, self.db.profile.trackerFontSize)
				dash.text:SetText(" - ")
				dash.text:SetTextColor(0.8, 0.8, 0.8)
				
				local yOffset = WATCHFRAMELINES_FONTSPACING
				if(lastLine.statusBar:IsShown()) then yOffset = yOffset - 5 end
				
				local dashWidth = dash.text:GetStringWidth()
				
				dash:SetWidth(dashWidth)
				dash:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", nextXOffset, yOffset)
				dash:Show()
				
				line = self:GetAchievementLine()
				line.text:SetFont(font, self.db.profile.trackerFontSize)
				line.text:SetText(description)
				line.text:SetTextColor(0.8, 0.8, 0.8)
				
				local stringWidth = line.text:GetStringWidth()
				local desiredWidth = math_ceil(stringWidth + dashWidth)
				if(desiredWidth > maxWidth and maxWidth < MAX_LONG_CRITERIA_WIDTH) then
					maxWidth = math_min(desiredWidth, MAX_LONG_CRITERIA_WIDTH)
				end
				local linesNeeded = math_ceil(stringWidth / (maxWidth - dashWidth))
				
				line:SetPoint("TOPLEFT", dash, "TOPRIGHT")
				line:SetPoint("TOPRIGHT", lastLine, "TOPRIGHT", 0, yOffset)
				line:SetHeight((WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING) + (WATCHFRAMELINES_FONTHEIGHT * (linesNeeded - 1)))
				line:Show()
				
				nextXOffset = -dashWidth
				lastLine = line
				
				for criteriaID, timedCriteria in next, WATCHFRAME_TIMEDCRITERIA do
					if(timedCriteria.achievementID == achievementID) then
						line = self:GetAchievementLine()
						line.criteriaID = criteriaID
						line.duration = timedCriteria.duration
						line.startTime = timedCriteria.startTime
						line.text:SetFont(font, self.db.profile.trackerFontSize)
						
						local yOffset = WATCHFRAMELINES_FONTSPACING
						if(lastLine.statusBar:IsShown()) then yOffset = yOffset - 5 end
						
						line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset)
						line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", nextXOffset, yOffset)
						line:Show()
						
						lastLine = line
						nextXOffset = 0
						WatchFrameLines_AddUpdateFunction(WatchFrame_UpdateTimedAchievements)
					end
				end
			end
		end
	end
	
	self:ReleaseUnusedAchievementLines()
	
	if(maxWidth > 0 and maxWidth < 194) then maxWidth = 194 end
	
	return heightUsed, maxWidth
end

function Inquest:DisplayQuestTimers(lineFrame, initialOffset, maxHeight, frameWidth, ...)
	local numTimers = select("#", ...)
	local maxWidth = 0
	local heightUsed = 0
	local lineHeight = WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING
	
	self:ResetTimerLines()
	
	local font = LSM:Fetch("font", self.db.profile.trackerFont)
	
	if(numTimers > 0) then
		local line = self:GetTimerLine()
		line.text:SetFont(font, self.db.profile.trackerFontSize)
		line.text:SetText(NORMAL_FONT_COLOR_CODE .. QUEST_TIMERS)
		line:Show()
		line:SetPoint("TOPLEFT", lineFrame, "TOPLEFT", 0, initialOffset)
		line:SetPoint("TOPRIGHT", lineFrame, "TOPRIGHT", 0, initialOffset)
		
		maxWidth = line.text:GetStringWidth()
		heightUsed = lineHeight
		local lastLine = line
		
		for i=1, numTimers do
			line = self:GetTimerLine()
			line.text:SetFont(font, self.db.profile.trackerFontSize)
			line.text:SetText(" - " .. SecondsToTime(select(i, ...)))
			line:Show()
			line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT")
			line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT")
			
			heightUsed = heightUsed + lineHeight
			maxWidth = math_max(maxWidth, line.text:GetStringWidth())
			line:SetWidth(frameWidth)
			
			line:SetScript("OnEnter", function(frame)
				GameTooltip:SetOwner(frame)
				GameTooltip:SetHyperlink(GetQuestLink(GetQuestIndexForTimer(i)))
				GameTooltip:Show()
			end)
			line:SetScript("OnLeave", GameTooltip_Hide)
			line:EnableMouse(true)
			
			lastLine = line
		end
		
		if(WATCHFRAME_NUM_TIMERS ~= numTimers) then
			WATCHFRAME_NUM_TIMERS = numTimers
			WatchFrameLines_AddUpdateFunction(WatchFrame_HandleQuestTimerUpdate)
		end
	elseif(WATCHFRAME_NUM_TIMERS ~= 0) then
		WatchFrameLines_RemoveUpdateFunction(WatchFrame_HandleQuestTimerUpdate)
		WATCHFRAME_NUM_TIMERS = 0
	end
	
	self:ReleaseUnusedTimerLines()
	
	return heightUsed, maxWidth
end
